/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "monitor.h"
#include "net.h"
#include "socket.h"


/**
 *  Methods for controlling services managed by monit.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Rory Toma, <rory@digeo.com>
 *  @author Martin Pala, <martin.pala@iol.cz>
 *
 *  @version \$Id: control.c,v 1.50 2003/06/26 14:24:11 hauk Exp $
 *
 *  @file
 */


/* -------------------------------------------------------------- Prototypes */


static void do_stop(Service_T);
static void do_start(Service_T);
static void wait_stop(Service_T);
static void wait_start(Service_T);
static void do_depend(Service_T, char *);


/* ------------------------------------------------------------------ Public */


/**
 * Start/stop all processes in service list
 * @param action A string describing the action to execute
 */
void control(char *action) {

  Service_T s;

  ASSERT(action);

  for(s= servicelist; s; s= s->next) {
    if(s->visited)
	continue;
    if(exist_daemon()) {
      d_check_service(s->name, action);
    } else {
      check_service(s->name, action);
    }
  }
  
  reset_depend();

}


/*
 * Start/stop all processes in a group
 * @param G group name
 * @param action A string describing the action to execute
 */
void control_group(char *G, char *action) {

  Service_T s;

  ASSERT(G);
  ASSERT(action);

  for(s= servicelist; s; s= s->next) {
    if(s->visited)
	continue;
    if(IS(s->group, G)) {
      if(exist_daemon()) {
	d_check_service(s->name, action);
      } else {
	check_service(s->name, action);
      }
    }
  }

  reset_depend();
 
}


/**
 * Pass on to methods in http/cervlet.c to start/stop services
 * @param P A service name as stated in the config file
 * @param action A string describing the action to execute
 */
void d_check_service(char *P, char *action) {

  Socket_T s;
  char *auth= get_basic_authentication_header();

  ASSERT(P);
  ASSERT(action);


  s= socket_new(Run.bind_addr?Run.bind_addr:"localhost",
		Run.httpdport, SOCK_STREAM, Run.httpdssl);
  
  if(!s) {
    
    log("%s: Cannot connect to the monit daemon. "
          "Did you start it with http support?\n", prog);
    goto error;
    
  } else {

    socket_print(s, "GET /%s?action=%s HTTP/1.0\r\n%s\r\n", P, action, auth);
    socket_free(&s);
      
  }
  
  error:
  free(auth);
  
}


/**
 * Check to see if we should try to start/stop service
 * @param P A service name as stated in the config file
 * @param action A string describing the action to execute
 */
void check_service(char *P, char *action) {

  Service_T s= NULL;

  ASSERT(P);
  ASSERT(action);

  if(NULL==(s= get_service(P))) {
    log("%s: Cannot %s program '%s' -- not found in %s\n",
          prog, action, P, Run.controlfile);
    return;
  }

  if(IS(action, "start")) {
    
    if(s->task==TASK_PROCESS && is_process_running(s)) {

      if(!s->do_validate) {

        LOCK(Run.mutex)
          s->do_validate= TRUE;
        END_LOCK;

        DEBUG("Monitoring enabled -- process %s\n", s->name);

      }

      return;
    }
    
    if(!s->start) {
      log("%s: Start method not defined -- process %s\n",
	    prog, P);
      return;
    }
    
    do_depend(s, "stop");
    do_start(s);
    do_depend(s, "start");
    
  } else if(IS(action, "stop")) {
    
    if(!s->stop) {
      log("%s: Stop method not defined -- process %s\n",
	    prog, P);
      return;
    }
    
    do_depend(s, "stop");
    do_stop(s);
    
  } else if(IS(action, "restart")) {
    
    if(!s->start || !s->stop) {
      log("%s: Start or stop method not defined -- process %s\n",
	    prog, P);
      return;
    }
    
    do_depend(s, "stop");
    do_stop(s);
    do_start(s);
    do_depend(s, "start");
    
  }
    
}


/*
 * Reset the visited flags used when handling dependencies
 */
void reset_depend() {

  Service_T s;
  
  for(s= servicelist; s; s= s->next) {
    s->visited= FALSE;
    s->depend_visited= FALSE;
  }

}


/* ----------------------------------------------------------------- Private */


/*
 * This is a post- fix recursive function for starting every service
 * that s depends on before starting s.
 * @param s A Service_T object
 */
static void do_start(Service_T s) {

  ASSERT(s);

  if(s->visited)
      return;
  
  s->visited= TRUE;
  
  if(s->dependantlist) {
    
    Dependant_T d;
    
    for(d= s->dependantlist; d; d= d->next ) {
      
      Service_T dp= get_service(d->dependant);
      ASSERT(dp);
      do_start(dp);
      
    }
  }
  
  if(s->start && (s->task!=TASK_PROCESS || !is_process_running(s))) {
    log("start: (%s) %s\n", s->name, s->start->arg[0]);
    spawn(s, s->start);
    if(s->task==TASK_PROCESS)
      wait_start(s);
  }
  
  LOCK(Run.mutex)
    s->do_validate= TRUE;
  END_LOCK;

  DEBUG("Monitoring enabled -- process %s\n", s->name);

}


/*
 * This function simply stops the service p.
 * @param s A Service_T object
 */
static void do_stop(Service_T s) {

  ASSERT(s);

  if(s->depend_visited)
      return;
  
  s->depend_visited= TRUE;
  
  LOCK(Run.mutex)
    s->do_validate= FALSE;
  END_LOCK;

  DEBUG("Monitoring disabled -- service %s\n", s->name);

  if(s->stop && (s->task!=TASK_PROCESS || is_process_running(s))) {
    log("stop: (%s) %s\n", s->name, s->stop->arg[0]);
    spawn(s, s->stop);
    if(s->task==TASK_PROCESS)
      wait_stop(s);
  }

  /* Reset the proc info object in case of a later restart */
  memset(s->procinfo, 0, sizeof *(s->procinfo));
  
}


/*
 * This is an in-fix recursive function called before s is started to
 * stop every service that depends on s, in reverse order *or* after s
 * was started to start again every service that depends on s. The
 * action parametere controls if this function should start or stop
 * the procceses that depends on s.
 * @param s A Service_T object
 * @param action An action to do on the dependant services
 */
static void do_depend(Service_T s, char *action) {

  Service_T parent;
  
  ASSERT(s);

  for(parent= servicelist; parent; parent= parent->next) {
    
    if(parent->dependantlist) {

      Dependant_T d;
    
      for(d= parent->dependantlist; d; d= d->next)
	  if(IS(d->dependant, s->name))
	     break;
      
      if(d) {
	
	if(IS(action, "start"))
	    do_start(parent);
	
	do_depend(parent, action);
	
	if(IS(action, "stop"))
	    do_stop(parent);
	
      }
    }
  }
}
    

/*
 * This function suspend the control until the service s is running.
 * @param s A Service to wait for
 */
static void wait_start(Service_T s) {

  int max_tries= Run.polltime;
  
  ASSERT(s);

  while(max_tries--) {
    if(is_process_running(s))
      break;
    sleep(1);
  }
  
  if(!is_process_running(s))
      log("%s: Warning process '%s' was not started\n", prog, s->name);
  
}


/*
 * This function suspend the control until the service s is stopped.
 * @param s A Service to wait for
 */
static void wait_stop(Service_T s) {

  int max_tries= Run.polltime;
  
  ASSERT(s);

  while(max_tries--) {
    if(!is_process_running(s))
      break;
    sleep(1);
  }

  if(is_process_running(s))
      log("%s: Warning process '%s' was not stopped\n", prog, s->name);

}
