/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_SYS_VFS_H
# include <sys/vfs.h>
#endif

#ifdef HAVE_SYS_STATVFS_H
# include <sys/statvfs.h>
#endif

#ifdef HAVE_MNTENT_H
#include <mntent.h>
#endif

#ifdef HAVE_SYS_MNTTAB_H
# include <sys/mnttab.h>
#endif


#include "monitor.h"


#if defined HAVE_STATVFS
#define statfs statvfs
#endif


/**
 *  Utilities used for managing files used by monit.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp, <chopp@iei.tu-clausthal.de>
 *  @author Martin Pala, <martin.pala@iol.cz>
 *
 *  @version \$Id: files.c,v 1.34 2003/06/16 22:06:15 hauk Exp $
 *
 *  @file
 */


/* ------------------------------------------------------------------ Public */


/**
 * Initialize the programs file variables
 */
void init_files() {

  char pidfile[STRLEN];
  char statefile[STRLEN];

  /* Check if the pidfile was already set during configfile parsing */
  if(Run.pidfile == NULL) {

    /* Set the location of this programs pidfile */
    if(! getuid()) {

      snprintf(pidfile, STRLEN, "%s/%s", MYPIDDIR, MYPIDFILE);

    } else {

      snprintf(pidfile, STRLEN, "%s/.%s", Run.Env.home, MYPIDFILE);

    }

    Run.pidfile= xstrdup(pidfile);

  }

  /* Set the location of the programs state file */
  if(Run.statefile == NULL) {

    snprintf(statefile, STRLEN, "%s/.%s", Run.Env.home, MYPSTATEFILE);
    
    Run.statefile= xstrdup(statefile);

  }
  

}


/**
 * Remove temporary files
 */
void finalize_files() {
  
  unlink(Run.pidfile);

}


/**
 * Get a object's last modified timestamp.
 * @param object A object to stat
 * @param type Requested object's type
 * @return Max of either st_mtime or st_ctime or
 * FALSE if not found or different type of object
 */
time_t get_timestamp(char *object, mode_t type) {
  
  struct stat buf;

  ASSERT(object);

  if(! stat(object, &buf)) {
    
    if(((type == S_IFREG) && S_ISREG(buf.st_mode)) ||
       ((type == S_IFDIR) && S_ISDIR(buf.st_mode)) ||
       ((type == (S_IFREG|S_IFDIR)) && (S_ISREG(buf.st_mode) ||
					S_ISDIR(buf.st_mode)))
       ) {
      
       return MAX(buf.st_mtime, buf.st_ctime);

     } else {

       log("%s: Invalid object type - %s\n", prog, object);

     }

  }

  return FALSE;
  
}


/**
 * Search the system for the monit control file. Try first
 * ~/.monitrc, if that fails try ./monitrc and finally /etc/monitrc.
 * Exit the application if the control file is not found.
 * @return The location of monits control file (monitrc)
 */
char *find_rcfile() {

  char *rcfile= xmalloc(STRLEN);
  
  snprintf(rcfile, STRLEN, "%s/.%s", Run.Env.home, MONITRC);
  
  if(exist_file(rcfile)) {
    
    return (rcfile);
    
  }

  if(exist_file(MONITRC)) {
    
    memset(rcfile, 0, STRLEN);
    snprintf(rcfile, STRLEN, "%s/%s", Run.Env.cwd, MONITRC);
    
    return (rcfile);
    
  }
  
  memset(rcfile, 0, STRLEN);
  snprintf(rcfile, STRLEN, "/etc/%s", MONITRC);
  
  if(exist_file(rcfile)) 
    return (rcfile);
  
  log("%s: Cannot find the control file at ~/.%s, ./%s or at /etc/%s\n",
      prog, MONITRC, MONITRC, MONITRC);
  
  exit(1);
  
}


/**
 * Create a program's pidfile - Such a file is created when in daemon
 * mode. The file is created with mask = MYPIDMASK (usually 644).  
 * @param pidfile The name of the pidfile to create
 * @return TRUE if the file was created, otherwise FALSE. 
 */
int create_pidfile(char *pidfile) {
  
  FILE *F= NULL;

  ASSERT(pidfile);
  
  umask(MYPIDMASK);
  
  if ((F= fopen(pidfile,"w")) == (FILE *)NULL) {
    
    log("%s: Error opening pidfile '%s' for writing -- %s\n",
	prog, pidfile, STRERROR);
    
    return(FALSE);
    
  }
  
  fprintf(F, "%d", (int)getpid());
  fclose(F);

  return TRUE;
  
}


/**
 * Secure check the monitrc file. The run control file must have the
 * same uid as the REAL uid of this process, it must have permissions
 * no greater than 700 and it must not be a symbolic link.  We check
 * these conditions here.
 * @param rcfile The monitrc file
 * @return TRUE if the test passed otherwise FALSE
 */
int check_rcfile(char *rcfile) {

  ASSERT(rcfile);
  
  return check_file(rcfile, "control file", S_IRUSR | S_IWUSR | S_IXUSR);
  
}


/**
 * Check if the file is a regular file
 * @param file A path to the file to check
 * @return TRUE if file exist and is a regular file, otherwise FALSE
 */
int isreg_file(char *file) {
  
  struct stat buf;
  
  ASSERT(file);

  return (stat(file, &buf) == 0 && S_ISREG(buf.st_mode));
  
}


/**
 * Check if the file exist on the system
 * @file A path to the file to check
 * @return TRUE if file exist otherwise FALSE
 */
int exist_file(char *file) {
  
  struct stat buf;
  
  ASSERT(file);

  return (stat(file, &buf) == 0);
  
}


/**
 * Security check for files. The files must have the same uid as the
 * REAL uid of this process, it must have permissions no greater than
 * "maxpermission" and it must not be a symbolic link.  We check these
 * conditions here.
 * @param filename The filename of the checked file
 * @param description The description of the checked file
 * @param permmask The permission mask for the file
 * @return TRUE if the test passed otherwise FALSE
 */
int check_file(char *filename, char *description, int permmask) {

  struct stat buf;
  errno= 0;

  ASSERT(filename);
  ASSERT(description);

  if(lstat(filename, &buf) < 0) {
    
    log("%s: Cannot stat the %s '%s' -- %s\n",
	prog, description, filename, STRERROR);

    return FALSE;
    
  }
    
  if(S_ISLNK(buf.st_mode)) {
    
    log("%s: The %s '%s' must not be a symbolic link.\n",
	prog, description, filename);
    
    return(FALSE);
    
  }

  if(!S_ISREG(buf.st_mode)) {
    
    log("%s: The %s '%s' is not a regular file.\n", 
	prog, description,  filename);
    
    return FALSE;

  }

  if(buf.st_uid != geteuid())  {
    
    log("%s: The %s '%s' must be owned by you.\n", 
	prog, description, filename);
	  
    return FALSE;
    
  }

  if((buf.st_mode & 0777 ) & ~permmask) {

    /* 
       Explanation: 

           buf.st_mode & 0777 ->  We just want to check the
                                  permissions not the file type... 
                                  we did it already!
           () & ~permmask ->      We check if there are any other
                                  permissions set than in permmask 
    */

    log("%s: The %s '%s' must have permissions no more "
	"than -%c%c%c%c%c%c%c%c%c (0%o); "
	"right now permissions are -%c%c%c%c%c%c%c%c%c (0%o).\n", 
	prog, description, filename, 
	permmask&S_IRUSR?'r':'-',
	permmask&S_IWUSR?'w':'-',
	permmask&S_IXUSR?'x':'-',
	permmask&S_IRGRP?'r':'-',
	permmask&S_IWGRP?'w':'-',
	permmask&S_IXGRP?'x':'-',
	permmask&S_IROTH?'r':'-',
	permmask&S_IWOTH?'w':'-',
	permmask&S_IXOTH?'x':'-',
	permmask&0777,
	buf.st_mode&S_IRUSR?'r':'-',
	buf.st_mode&S_IWUSR?'w':'-',
	buf.st_mode&S_IXUSR?'x':'-',
	buf.st_mode&S_IRGRP?'r':'-',
	buf.st_mode&S_IWGRP?'w':'-',
	buf.st_mode&S_IXGRP?'x':'-',
	buf.st_mode&S_IROTH?'r':'-',
	buf.st_mode&S_IWOTH?'w':'-',
	buf.st_mode&S_IXOTH?'x':'-',
	buf.st_mode& 0777);
    
    return FALSE;
	
  }

  return TRUE;

}


/**
 * This function validates if given object is valid for filesystem
 * informations statistics. Valid objects are file or directory that
 * is part of requested filesystem or block special device. In the
 * case of file or directory the result is original object, in the
 * case of block special device mountpoint is returned. In any case,
 * filesystem must be mounted.
 * @param path String to store result
 * @param path_length Defines max. length of the resulting string
 * @param object Identifies appropriate device object
 * @return NULL in the case of failure otherwise pointer to
 * resulting path
 */
char *get_device(char *path, unsigned path_length, char *object) { 

  struct stat buf;

  ASSERT(path && object);

  if(stat(object, &buf) != 0) {
    log("%s: Cannot stat '%s' -- %s\n", prog, object, STRERROR);
    return NULL;
  }

  memset(path, '\0', path_length);

  if(S_ISREG(buf.st_mode) || S_ISDIR(buf.st_mode)) {

    return strncpy(path, object, path_length-1);

  } else if(S_ISBLK(buf.st_mode)) {

    FILE *mntfd;

#ifdef HAVE_MTAB

    struct mntent *mnt;

    if((mntfd= setmntent("/etc/mtab", "r")) == NULL) {
      log("%s: Cannot open /etc/mtab file", prog);
      return NULL;
    }

    /* Last match is significant */
    while((mnt= getmntent(mntfd)) != NULL) {

      if(IS(object, mnt->mnt_fsname))
        strncpy(path, mnt->mnt_dir, path_length-1);

    }

    endmntent(mntfd);

#elif defined HAVE_MNTTAB

    struct mnttab mnt;

    if((mntfd= fopen("/etc/mnttab", "r")) == NULL) {
      log("%s: Cannot open /etc/mnttab file", prog);
      return NULL;
    }

    /* Last match is significant */
    while(getmntent(mntfd, &mnt) == 0) {

      if(IS(object, mnt.mnt_special))
        strncpy(path, mnt.mnt_mountp, path_length-1);

    }

    fclose(mntfd);

#else

    log("%s: Unsupported mounted filesystem information method", prog);
    return NULL;

#endif

    return path;

  }

  log("%s: Not file, directory or block special device: '%s'", prog, object);

  return NULL;

}


/**
 * Filesystem usage statistics
 * @param object Identifies requested device
 * @param buf Information structure where resulting data will be stored
 * @return TRUE if informations were succesfully read otherwise FALSE
 */
int get_fsusage(char *object, DeviceInfo_T buf) { 

#if (defined HAVE_STATFS) || (defined HAVE_STATVFS)

  struct statfs usage;

#else

  log("%s: Unsupported filesystem informations gathering method\n", prog);

  return FALSE;

#endif

  char path[STRLEN];

  ASSERT(object);
  ASSERT(buf);

  if(!get_device(path, STRLEN, object)) {
    log("%s: Error looking for filesystem '%s'\n", prog, object);
    return FALSE;
  }

  if(statfs(path, &usage) != 0) {
    log("%s: Error getting usage statistics for device '%s' -- %s\n",
	prog, object, STRERROR);
    return FALSE;
  }

  buf->f_bsize=           usage.f_bsize;
  buf->f_blocks=          usage.f_blocks;
  buf->f_blocksfree=      usage.f_bavail;
  buf->f_blocksfreetotal= usage.f_bfree;
  buf->f_files=           usage.f_files;
  buf->f_filesfree=       usage.f_ffree;

  return TRUE;

}

