/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#ifndef MONITOR_H
#define MONITOR_H

#include <config.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include <signal.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>

#include "ssl.h"

#define VERSION            PACKAGE_VERSION
#define MONITRC            "monitrc"
#define TIMEFORMAT         "%Z %b %e %T"
#define STRERROR            strerror(errno)
#define STRLEN             256
#define ARGMAX             64
/* Set log file mode: "-rw-rw-r--" */
#define LOGMASK            0112
/* Set pid file mode: "-rw-r--r--" */
#define MYPIDMASK          0122
#define MYPIDDIR           PIDDIR
#define MYPIDFILE          "monit.pid"
#define MYPSTATEFILE       "monit.state"

#define LOCALHOST          "localhost"
#define SMTP_PORT          25

#define SMTP_TIMEOUT       30
#define CHECK_TIMEOUT      30
#define SSL_TIMEOUT        15

#define START_HTTP         1
#define STOP_HTTP          2

#define TRUE               1
#define FALSE              0

#define MODE_ACTIVE        0
#define MODE_PASSIVE       1
#define MODE_MANUAL        2

#define OPERATOR_GREATER   0
#define OPERATOR_LESS      1
#define OPERATOR_EQUAL     2
#define OPERATOR_NOTEQUAL  3

#define TIME_SECOND	   1
#define TIME_MINUTE	   60
#define TIME_HOUR	   3600
#define TIME_DAY	   86400

#define ACTION_ALERT       1
#define ACTION_RESTART     2
#define ACTION_STOP        3

#define TASK_DEVICE        0
#define TASK_DIRECTORY     1
#define TASK_FILE          2
#define TASK_PROCESS       3

#define RESOURCE_ID_CPU_PERCENT       1
#define RESOURCE_ID_MEM_PERCENT       2
#define RESOURCE_ID_MEM_KBYTE         3
#define RESOURCE_ID_LOAD1             4
#define RESOURCE_ID_LOAD5             5
#define RESOURCE_ID_LOAD15            6
#define RESOURCE_ID_CHILDREN          7
#define RESOURCE_ID_TOTAL_MEM_KBYTE   8
#define RESOURCE_ID_TOTAL_MEM_PERCENT 9
#define RESOURCE_ID_INODE             10
#define RESOURCE_ID_SPACE             11

#define UNIT_BYTE          1
#define UNIT_KILOBYTE      1024
#define UNIT_MEGABYTE      1048580
#define UNIT_GIGABYTE      1073740000


/** ------------------------------------------------- Special purpose macros */


/* Mask compiler built-in namespace (C99) and reserved identifiers as
 *  well as external identifiers (in libraries we link with). */
#define log  log_log

/* Replace the standard signal function with a more reliable using
 * sigaction. Taken from Stevens APUE book. */
typedef void Sigfunc(int);
Sigfunc *signal(int signo, Sigfunc * func);
#if defined(SIG_IGN) && !defined(SIG_ERR)
#define SIG_ERR ((Sigfunc *)-1)
#endif


/** ------------------------------------------------- General purpose macros */


#ifdef MAX
#undef MAX
#endif
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#define IS(a,b)	 ((a&&b)?!strcasecmp(a, b):0)
#define DEBUG if(Run.debug) log


/** ------------------------------------------------- Synchronization macros */


#define LOCK(mutex) do { pthread_mutex_t *_yymutex = &(mutex); \
        pthread_mutex_lock(_yymutex);
#define END_LOCK pthread_mutex_unlock(_yymutex); } while (0)


/** ----------------------------------------------- Object Constructor macro */


#define NEW(p) (xcalloc(1, (long)sizeof *(p)))


/** ------------------------------------------ Simple Assert Exception macro */


#define ASSERT(e) if(!(e)) { log("AssertException: at %s:%d\naborting..\n", \
    __FILE__, __LINE__); abort(); }


/* --------------------------------------------------------- Data structures */


/** Defines data for application runtime */
struct myrun {
  char *controlfile;                /**< The file to read configuration from */
  char *logfile;                         /**< The file to write logdata into */
  char *localhostname;                      /**< The host name for localhost */
  char *mailserver;              /**< Mailserver used for alert notification */
  char *pidfile;                                  /**< This programs pidfile */
  char *statefile;                /**< The file with the saved runtime state */
  char *mygroup;                              /**< Group Name of the Service */
  int  have_tty;                                 /**< Do we have a valid tty */
  int  debug;                   /**< Write debug information - TRUE or FALSE */
  int  use_syslog;                          /**< If TRUE write log to syslog */
  int  dolog;       /**< TRUE if program should log actions, otherwise FALSE */
  int  isdaemon;                 /**< TRUE if program should run as a daemon */
  int  polltime;        /**< In deamon mode, the sleeptime (sec) between run */
  int  dohttpd;                    /**< TRUE if monit HTTP server should run */
  int  httpdssl;                     /**< TRUE if monit HTTP server uses ssl */
  char *httpsslpem;                       /**< PEM file for the HTTPS server */
  int  clientssl;   /**< TRUE if monit HTTP server uses ssl with client auth */
  char *httpsslclientpem;      /**< PEM file/dir to check against at connect */
  int  servicessl;                 /**< TRUE if monit has ssl service checks */
  int  allowselfcert;   /**< TRUE if self certified client certs are allowed */
  int  httpdport;                    /**< The monit http server's portnumber */
  int  validate;	       /**< TRUE - use validate behaviour on startup */
  int  init; 		       /**< TRUE - don't background to run from init */
  int  facility;              /** The facility to use when running openlog() */
  int  doprocess;                 /**< TRUE if process status engine is used */
  char *bind_addr;                  /**< The address monit http will bind to */
  int  doreload;             /**< TRUE if a monit daemon should reinitialize */
  mode_t umask;                /**< The initial umask monit was started with */
  int  testing;   /**< Running in configuration testing mode - TRUE or FALSE */

  double loadavg[3];                                /**< Load average triple */

       /** An object holding program relevant "environment" data, see; env.c */
  struct myenvironment {
    char *user;             /**< The the effective user running this program */
    char *home;                                    /**< Users home directory */
    char *cwd;                                /**< Current working directory */
  } Env;
                      /** An object holding Basic Authentication information */
  struct myautentication {
    int  defined;              /**< TRUE if basic authentication can be used */
    char *uname;                 /**< User allowed to connect to monit httpd */
    char *passwd;                      /**< The users password in clear text */
  } Auth;
                                      /** User selected standard mail format */
  struct myformat {
    char *from;                          /**< The standard mail from address */
    char *subject;                            /**< The standard mail subject */
    char *message;                            /**< The standard mail message */
  } MailFormat;
                                          
  pthread_mutex_t mutex;    /**< Mutex used for service data synchronization */

};


/** Defines a port object */
typedef struct myport {
  volatile int socket;                       /**< Socket used for connection */
  int  type;                  /**< Socket type used for connection (UDP/TCP) */
  int  family;            /**< Socket family used for connection (INET/UNIX) */
  char *hostname;                                     /**< Hostname to check */
  int  port;                                                 /**< Portnumber */
  char *request;                              /**< Specific protocol request */
  char *pathname;                   /**< Pathname, in case of an UNIX socket */
  char *address;               /**< Human readable destination of the socket */

  ssl_connection * ssl;
  char *certmd5;
  /**< Object used for testing a port's service */
  struct myprotocol {
    const char *name;                                     /**< Protocol name */
    int(*check)(struct myport *);         /**< Service verification function */
  } *protocol;
  
  /** For internal use */
  struct myport *next;                               /**< next port in chain */
} *Port_T;


/** Defines a protocol object with protocol functions */
typedef struct myprotocol *Protocol_T;


typedef struct mydependant {
  char *dependant;                           /**< name of dependant service  */
  /** For internal use */
  struct mydependant *next;           /**< next dependant service in chain   */
} *Dependant_T;


/** Defines a checksum object */
typedef struct mychecksum {
  char *file;                  /**< A filename to compute a md5 checksum for */
  char *md5;                       /**< A md5 checksum computed for the file */

  /** For internal use */
  struct mychecksum *next;                       /**< next checksum in chain */
} *Checksum_T;


/** Defines an alert object */
typedef struct myalert {
  int  alert_on_timeout;       /**< If TRUE, alert user when service timeout */
  int  alert_on_restart;  /**< If TRUE, alert user when the service restarts */
  int  alert_on_checksum;    /**< If TRUE, alert user when the checksum fail */
  int  alert_on_resource;     /**< If TRUE, alert user when resources exceed */
  int  alert_on_stop;          /**< If TRUE, alert user when service stopped */
  int  alert_on_timestamp;  /**< If TRUE, alert user when the timestamp fail */
} Alert_T;


/** Defines a mailinglist object */
typedef struct mymail {
  char *to;                         /**< Mail address for alert notification */
  char *from;                                     /**< The mail from address */
  char *subject;                                       /**< The mail subject */
  char *message;                                       /**< The mail message */
  char *opt_message;                 /**< An optional message used in alerts */
  Alert_T alerts; /*< Alert events for which this mail object should be sent */

  /** For internal use */
  struct mymail *next;                          /**< next recipient in chain */
} *Mail_T;


/** Defines resource data */
typedef struct myresource {
  int  resource_id;                              /**< Which value is checked */
  long limit;                                     /**< Limit of the resource */
  int  operator;                                    /**< Comparison operator */
  int  cycle;                                     /**< Cycle overrun counter */
  int  max_cycle;                                   /**< Cycle overrun limit */
  int  action;                                /**< Action in case of failure */

  /** For internal use */
  struct myresource *next;                       /**< next resource in chain */
} *Resource_T;


/** Defines timestamp object */
typedef struct mytimestamp {
  char *pathname;				     /**< Path to the object */
  int  operator;				    /**< Comparison operator */
  int  time;					    /**< Timestamp watermark */
  int  action;				      /**< Action in case of failure */

  /** For internal use */
  struct mytimestamp *next;                     /**< next timestamp in chain */
} *Timestamp_T;


/** Defines procfs (or other mechanism) data */
typedef struct myprocinfo {
  int  pid;
  int  ppid;
  int  status_flag;
  long mem_kbyte;    
  int  mem_percent;                                      /**< pecentage * 10 */
  int  cpu_percent;                                      /**< pecentage * 10 */
  int  children;
  long total_mem_kbyte;
  int  total_mem_percent;                                /**< pecentage * 10 */

  /* private for calculating cpu_percent */
  double time_prev;                                        /**< 1/10 seconds */
  long  cputime;                                           /**< 1/10 seconds */
  long  cputime_prev;                                      /**< 1/10 seconds */
} *ProcInfo_T;


/**
 * Defines a Command with ARGMAX optional arguments. The arguments
 * array must be NULL terminated and the first entry is the program
 * itself. In addition, a user and group may be set for the Command
 * which means that the Command should run as a certain user and with
 * certain group.  Finally, a Command_T object may contain a set of
 * environment variables to be set before the Command is started.
 */
typedef struct mycommand {
  char *arg[ARGMAX];                             /**< Program with arguments */
  int length;                         /**< The length of the arguments array */
  int has_uid;            /**< TRUE if a new uid is defined for this Command */
  uid_t uid;         /**< The user id to switch to when running this Command */
  int has_gid;            /**< TRUE if a new gid is defined for this Command */
  gid_t gid;        /**< The group id to switch to when running this Command */
} *Command_T;


/** Defines an alien program to be execute upon alert or upon an action. */
typedef struct myexec {
  Alert_T alerts;   /**< Alert events triggering this program to be executed */
  Command_T program;                       /**< The external program to call */
  
  /** For internal use */
  struct myexec *next;                   /**< next external program in chain */
} *Exec_T;


/** Defines filesystem device data */
typedef struct mydeviceinfo {
  long f_bsize;                                     /**< Transfer block size */
  long f_blocks;                        /**< Total data blocks in filesystem */
  long f_blocksfree;             /**< Free blocks available to non-superuser */
  long f_blocksfreetotal;                     /**< Free blocks in filesystem */
  long f_files;                          /**< Total file nodes in filesystem */
  long f_filesfree;                       /**< Free file nodes in filesystem */
} *DeviceInfo_T;


/** Defines filesystem device configuration */
typedef struct mydevice {
  int  resource;                        /**< Whether to check inode or space */
  int  operator;                                    /**< Comparison operator */
  long limit_absolute;                               /**< Watermark - blocks */
  int  limit_percent;                               /**< Watermark - percent */
  int  action;                                /**< Action in case of failure */

  /** For internal use */
  struct mydevice *next;                           /**< next device in chain */
} *Device_T;


/** Defines service data */
typedef struct myservice {

  /** Common parameters */
  int task;                                         /**< Monitored task type */
  char *name;                                  /**< Service descriptive name */
  char *group;                                       /**< Service group name */
  Command_T start;                    /**< The start command for the service */
  Command_T stop;                      /**< The stop command for the service */
  int  do_validate;    /**< validation flag, if FALSE, no validation is done */
  int  mode;                            /**< Monitoring mode for the service */
  int  ncycle;                          /**< The number of the current cycle */
  int  nstart;           /**< The number of current starts with this service */
  int  to_start;                                  /**< Timeout start ceiling */
  int  to_cycle;                                  /**< Timeout cycle ceiling */
  int  every;                        /**< Check this program at given cycles */
  int  nevery;          /**< Counter for every.  When nevery == every, check */
  int  has_checksum_error;     /**< TRUE if the service has a checksum error */
  int  def_every;              /**< TRUE if every is defined for the service */
  int  def_timeout;          /**< TRUE if timeout is defined for the service */
  int  def_checksum;        /**< TRUE if checksum is defined for the service */
  int  def_procinfo;        /**< TRUE if procinfo is defined for the service */
  int  visited;      /**< Service visited flag, set if dependencies are used */
  int  depend_visited;/**< Depend visited flag, set if dependencies are used */
  int  perm;                                          /**< Access permission */
  int  uid;                                                 /**< Owner's uid */
  int  gid;                                                 /**< Owner's gid */

  Port_T portlist;                   /**< Portnumbers the service listens on */
  Checksum_T checksumlist;          /**< A list of file associated checksums */
  Mail_T maillist;                       /**< Alert notification mailinglist */
  Resource_T resourcelist;                           /**< Resouce check list */
  Dependant_T dependantlist;                      /**<Dependant service list */
  Timestamp_T timestamplist;                       /**< Timestamp check list */
  Device_T devicelist;                                /**< Device check list */
  Exec_T execlist;    /**< External programs to call upon an alert or action */
  
  /** Runtime parameters */
  DeviceInfo_T devinfo;                       /**< Data for the device check */
  ProcInfo_T procinfo;                        /**< Data for the procfs check */

  /** Context specific parameters */
  char *path;   /**< Path to the device, file, directory or process pid file */
  
  /** For internal use */
  pthread_mutex_t mutex;          /**< Mutex used for action synchronization */
  struct myservice *next;                         /**< next service in chain */
  struct myservice *next_depend;           /**< next depend service in chain */

} *Service_T;


/** Defines process tree */
typedef struct myprocesstree {
  int                       pid;
  int                       ppid;
  long                      mem_kbyte;
  int                       children_num;
  long                      mem_kbyte_sum;
  int                       children_sum;

  /** For internal use */
  struct myprocesstree **   children;
  struct myprocesstree *    parent;
} ProcessTree_T;


/* -------------------------------------------------------- Global variables */


char   *prog;                                  /**< The Name of this Program */
struct myrun Run;                      /**< Struct holding runtime constants */
Service_T servicelist;                /**< The service list (created in p.y) */


/* ------------------------------------------------------- Public prototypes */


char *trim(char *);
char *ltrim(char *);
char *rtrim(char *);
char *trim_quotes(char *);
int   starts_with(char *,char *);
void  handle_string_escapes(char *);
void  chomp(char *);
int   exist_service(char *);
int   get_service_list_length();
Service_T get_service(char *);
void  printrunlist();
void  printservice(Service_T);
void  printservicelist();
pid_t get_pid(char *);
int   is_process_running(Service_T);
int   is_strdefined(char *);
int   isreg_file(char *);
char *stripfilename(char*);
int   exist_file(char *);
char *get_ctime();
char *get_RFC822date(long *);
char *get_uptime(long delta);
char *get_process_uptime(char *);
char *get_md5sum(char *);
int   check_md5(char *, char *);
char *url_encode(char *uri);
char *get_basic_authentication_header();
int   set_md5sum(char **, char *);
int   parse(char *);
void  control(char *);
void  control_group(char *, char *);
void  d_check_service(char *, char *);
void  check_service(char *, char *);
void  setup_dependants();
void  reset_depend();
void  spawn(Service_T, Command_T);
void  status();
void  status_group(char *);
int   log_init() ;
void  log_log(const char *format, ...) ;
void  log_close();
void  validate() ;
void  daemonize() ;
void  gc();
void  gc_service(Service_T*);
void  gc_mail_list(Mail_T*);
void  init_files();
time_t get_timestamp(char *, mode_t);
void  finalize_files();
char *find_rcfile();    
int   create_pidfile(char *);
int   check_rcfile(char *);
int   kill_daemon(int);
int   exist_daemon(); 
void  sendmail(Mail_T);
int   sock_msg(int sock, char *, ...);
void  init_env();
void *xmalloc (int);
void *xcalloc(long, long);
char *xstrdup(const char *);
void *xresize(void *, long);
void  set_alarm_handler(void *);
void  monit_http(int);
int   check_httpd();
int   can_http();
char *format(const char *, va_list);
void  redirect_stdfd();
pid_t getpgid(pid_t);
int   check_file(char *, char *, int);
void  state_save();
int   state_should_update();
void  state_update();
void  state_finalize();
char *get_device(char *, unsigned, char *);
int   get_fsusage(char *, DeviceInfo_T);

#endif
