/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


%{
  
/*
 * DESCRIPTION
 *   Simple context-free grammar for parsing the control file. 
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Olivier Beyssac, <ob@r14.freenix.oraddmailg>
 *  @author Kianusch Sayah Karadji <kianusch.sayah.karadji@sk-tech.net>
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *  @author Rory Toma <rory@digeo.com>
 *  @version \$Id: p.y,v 1.80 2003/06/24 02:49:09 hauk Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
  
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
  
#include <sys/socket.h>

#ifdef HAVE_ASM_PARAM_H
#include <asm/param.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#include "net.h"
#include "monitor.h"
#include "protocol.h"
#include "engine.h"
#include "alert.h"
#include "monit_process.h"
#include "ssl.h"

#define YERROR(msg) { yyerror(msg); YYABORT; }

  struct IHavePrecedence {
    int daemon;
    int logfile;
    int pidfile;
  }; 


  struct PortSet {
    int  socket;
    char *hostname;
    int  port;
    int  type;
    int  family;
    int  ssl;
    int  sslversion;
    char *request;
    char *pathname;
    char *certmd5;
    Protocol_T protocol;
  };

  struct ResourceSet {
    int  resource_id;
    int  limit;
    int  operator;
    int  max_cycle;
    int  action;
  };

  struct TimestampSet {
    char *pathname;
    int  operator;
    int  time;
    int  action;
  };

  struct DeviceSet {
    int  resource;
    int  operator;
    long limit_absolute;
    int  limit_percent;
    int  action;
  };

  /* yacc interface */
  void  yyerror(const char *);

  /* lexer interface */
  int yylex(void);
  extern FILE *yyin;
  extern int lineno;
  extern char *yytext;

  
  /* Local variables */
  static int cfg_errflag= FALSE;
  static Service_T tail= NULL;
  static Service_T current= NULL;
  static struct mymail mailset;
  static struct myalert alertset;
  static Command_T command= NULL;
  static Service_T depend_list= NULL;
  static struct IHavePrecedence ihp= {FALSE, FALSE, FALSE};
  static struct PortSet portset= {-1, NULL, 0, SOCK_STREAM, AF_INET, FALSE,
    SSL_VERSION_AUTO, NULL, NULL, NULL, NULL};
  static struct ResourceSet resourceset= {0, 0, OPERATOR_EQUAL, 1,
    ACTION_ALERT};
  static struct TimestampSet timestampset= {NULL, OPERATOR_EQUAL, 0,
    ACTION_ALERT};
  static struct DeviceSet deviceset= {0, OPERATOR_EQUAL, -1, -1, ACTION_ALERT};

  /* Private prototypes */
  static void initialize();
  static void addservice(Service_T);
  static void addmail(char *, Mail_T, Alert_T);
  static void createservice(int, char *, char *);
  static void adddependant(char *);
  static void addchecksum(char *, char *);
  static void addport(struct PortSet *);
  static void addresource(struct ResourceSet *);
  static void addtimestamp(struct TimestampSet *);
  static void adddevice(struct DeviceSet *);
  static void *addprotocol(int);
  static void addcommand(int);
  static void addargument(char *);
  static void adduid(char *user, uid_t uid);
  static void addgid(char *group, gid_t gid);
  static void setlogfile(char *);
  static void setpidfile(char *);
  static void setmailserver(char *);
  static void setuname(char *);
  static void setpasswd(char *);
  static void reset_mailset();
  static void reset_runmail();
  static void reset_portset();
  static void reset_resourceset();
  static void reset_timestampset();
  static void reset_deviceset();
  static void check_name(char *);
  static void check_timeout(int, int);
  static void check_every(int);
  static int  check_perm(int);
  static void check_hostname (char *);
  static void createinfo();
  static char *append_hostname(char *);
  static void cleanup_md5_string(char *);
  static void check_depend();
  static void order_depend(Service_T);
  static void validate_depend(Service_T, int *);
  static void setsyslog(char *);
%}

%union {
  float real;
  int   number;
  char *string;
}

%token SET LOGFILE DAEMON SYSLOG MAILSERVER HTTPD ALLOW ADDRESS INIT
%token PEMFILE ENABLE DISABLE HTTPDSSL CLIENTPEMFILE ALLOWSELFCERTIFICATION
%token STATEFILE
%token CHECK PIDFILE START STOP PATHTOK
%token HOST PORT TYPE UDP TCP TCPSSL PROTOCOL FACILITY
%token ALERT MAILFORMAT UNIXSOCKET
%token TIMEOUT RESTART CHECKSUM EXPECT EVERY 
%token DEFAULT HTTP FTP SMTP POP IMAP NNTP SSH DWP LDAP2 LDAP3 CERTMD5
%token <string> STRING PATH MAILADDR MAILFROM MAILSUBJECT 
%token <string> MAILBODY SERVICENAME
%token <number> NUMBER PERCENT
%token <real> REAL
%token CHECKPROC CHECKDEV CHECKFILE CHECKDIR 
%token CPUUSAGE MEMUSAGE MEMKBYTE TOTALMEMKBYTE TOTALMEMUSAGE CHILDREN 
%token RESOURCE MEMORY TOTALMEMORY CPU LOADAVG1 LOADAVG5 LOADAVG15 
%token AUTOSTART MODE ACTIVE PASSIVE MANUAL
%token GROUP REQUEST DEPENDS
%token UID GID ENVIRONMENT
%token TIMESTAMP SECOND MINUTE HOUR DAY
%token SSLAUTO SSLV2 SSLV3 TLSV1
%token BYTE KILOBYTE MEGABYTE GIGABYTE
%token INODE SPACE PERM

%left GREATER LESS EQUAL NOTEQUAL

%%
cfgfile         : /* EMPTY */
                | statement_list
                ;

statement_list  : statement
                | statement_list statement
                ;

statement       : setdaemon
                | setlog
                | setmta
                | setmailformat
                | sethttpd
                | setpid
                | setstate
                | setinit
                | checkproc optproclist
                | checkfile optfilelist
                | checkdev optdevlist
                | checkdir optdirlist
                ;


optproclist     : /* EMPTY */
                | optproclist optproc
                ;

optproc         : start
                | stop
                | host port type protocol {
                   addport(&portset);
                  }
                | unixsocket type protocol {
                   addport(&portset);
                  }
		| timestamp {
		   addtimestamp(&timestampset);
		  }
                | timeout
                | every
                | alert
		| checksum
		| autostart
		| mode
		| group
                | depend
		| resource resourcecycle action {
                   resourceset.action= $<number>3;
		   addresource(&resourceset);
		  }
                ;

optfilelist      : /* EMPTY */
                | optfilelist optfile
                ;

optfile         : start
                | stop
		| timestamp {
		   addtimestamp(&timestampset);
		  }
                | timeout
                | every
                | alert
		| perm
		| checksum
		| autostart
		| mode
		| group
                | depend
                ;

optdevlist      : /* EMPTY */
                | optdevlist optdev
                ;

optdev          : start
                | stop
                | timeout
                | every
                | alert
		| autostart
		| mode
		| group
                | depend
                | inode {
                   adddevice(&deviceset);
                  }
                | space {
                   adddevice(&deviceset);
                  }
                ;

optdirlist      : /* EMPTY */
                | optdirlist optdir
                ;

optdir          : start
                | stop
		| timestamp {
		   addtimestamp(&timestampset);
		  }
                | timeout
                | every
                | alert
		| perm
		| checksum
		| autostart
		| mode
		| group
                | depend
                ;

setdaemon       : SET DAEMON NUMBER  {
                   if(!Run.isdaemon || ihp.daemon) {
		     ihp.daemon= TRUE;
		     Run.isdaemon= TRUE;
		     Run.polltime= $3;
		   }
                 }
                ;

setinit         : SET INIT {
                   Run.init= TRUE;
                 }
                ;

setlog          : SET LOGFILE PATH   {
                   if(!Run.logfile || ihp.logfile) {
		     ihp.logfile= TRUE;
		     setlogfile($3);
		     Run.use_syslog= FALSE;
		     Run.dolog=TRUE;
		   }
                 }
                | SET LOGFILE SYSLOG { setsyslog(NULL); }
                | SET LOGFILE SYSLOG FACILITY STRING {
		    setsyslog($5); free($5);
		  }
		;

setstate        : SET STATEFILE PATH {
                    Run.statefile= $3;
                  }
                ;

setpid          : SET PIDFILE PATH {
                   if(!Run.pidfile || ihp.pidfile) {
		     ihp.pidfile= TRUE;
                     setpidfile($3);
		   }
                 }
                ;

setmta          : SET MAILSERVER STRING   {
                   check_hostname($3);
		   setmailserver($3);
                 }
                ;

setmailformat   : SET MAILFORMAT '{' formatoptionlist '}' {
                   Run.MailFormat.from=
			mailset.from?
			mailset.from:
			xstrdup(ALERT_FROM);
                   Run.MailFormat.subject=
			mailset.subject?
			mailset.subject:
			xstrdup(ALERT_SUBJECT);
                   Run.MailFormat.message=
			mailset.message?
			mailset.message:
			xstrdup(ALERT_MESSAGE);
		   reset_mailset();
                 }
                ;

sethttpd        : SET HTTPD PORT NUMBER httpdlist {
                   Run.dohttpd= TRUE;
		   Run.httpdport= $4;
                 }
                ;

httpdlist	: /* EMPTY */
                | httpdlist httpdoption
                ; 

httpdoption     : ssl
		| bindaddress
                | allow
                ;

bindaddress     : ADDRESS STRING { Run.bind_addr= $2; }
                ;

ssl             : ssldisable { Run.httpdssl= FALSE; }
                | sslenable pemfile clientpemfile allowselfcert { 
		  Run.httpdssl= TRUE;                   
		  if(!have_ssl()) {
		    YERROR("SSL is not supported\n");
		  }
		 }
                ;

sslenable       : HTTPDSSL
                | HTTPDSSL ENABLE
                | ENABLE HTTPDSSL
                ;

ssldisable      : HTTPDSSL DISABLE
                | DISABLE HTTPDSSL
                | ssldisable PEMFILE PATH
                | ssldisable CLIENTPEMFILE PATH
                | ssldisable ALLOWSELFCERTIFICATION
                ;

pemfile         : PEMFILE PATH {
                     Run.httpsslpem= $2;
                     if(!check_file(Run.httpsslpem,
				    "SSL server PEM file", S_IRWXU))
			 YERROR("SSL server PEM file");
                  }
                ;

clientpemfile   : /* EMPTY */ 
                | CLIENTPEMFILE PATH {
                     Run.httpsslclientpem= $2; 
		     Run.clientssl= TRUE;
                     if(!check_file(Run.httpsslclientpem,
				    "SSL client PEM file", 
				    S_IRWXU | S_IRGRP | S_IROTH))
			 YERROR("SSL client PEM file");
                  }
                ;

allowselfcert   : /* EMPTY */ { 
                   Run.allowselfcert= FALSE; 
                   config_ssl(Run.allowselfcert); 
                 }
                | ALLOWSELFCERTIFICATION {   
                   Run.allowselfcert= TRUE;
                   config_ssl(Run.allowselfcert); 
                 }
                ;

allow           : ALLOW STRING':'STRING {
                   setuname($2);
		   setpasswd($4);
		   Run.Auth.defined= TRUE;
                 }
                | ALLOW STRING {
		   if(!add_host_allow($2)) {
		     log("%s: Hostname '%s' did not resolve at line %d\n",
			 prog, $2, lineno-1);
		     cfg_errflag++;
		   }
		   free($2);
		 }
                ;

checkproc       : CHECKPROC SERVICENAME PIDFILE PATH {
                   createservice(TASK_PROCESS, $<string>2,$4);
		   check_name($<string>2);
                  }
                | CHECKPROC SERVICENAME PATHTOK PATH {
                   createservice(TASK_PROCESS, $<string>2,$4);
		   check_name($<string>2);
                  }
                ;

checkfile       : CHECKFILE SERVICENAME PATHTOK PATH {
                   createservice(TASK_FILE, $<string>2,$4);
		   check_name($<string>2);
                  }
                ;

checkdev        : CHECKDEV SERVICENAME PATHTOK PATH {
                   createservice(TASK_DEVICE, $<string>2,$4);
		   check_name($<string>2);
                  }
                ;

checkdir        : CHECKDIR SERVICENAME PATHTOK PATH {
                   createservice(TASK_DIRECTORY, $<string>2,$4);
		   check_name($<string>2);
                  }
                ;

start           : START argumentlist { addcommand(START); }
                | START argumentlist useroptionlist { addcommand(START); }
                ;

stop            : STOP argumentlist { addcommand(STOP); }
                | STOP argumentlist useroptionlist { addcommand(STOP); }
                ;

argumentlist    : argument
                | argumentlist argument
                ;

useroptionlist  : useroption
                | useroptionlist useroption
                ;

argument        : STRING { addargument($1); } | PATH { addargument($1); }
                ;

useroption      : UID STRING { adduid($2, 0); free($2); }
                | GID STRING { addgid($2, 0); free($2); }
                | UID NUMBER { adduid(NULL, $2); }
                | GID NUMBER { addgid(NULL, $2); }
                ;

host            : /* EMPTY */ { portset.hostname= xstrdup(LOCALHOST); }
                | HOST STRING { check_hostname($2); portset.hostname= $2; }
		;

port            : PORT NUMBER { portset.port= $2; portset.family= AF_INET; }
                ;

unixsocket      : UNIXSOCKET PATH {
                    portset.pathname= $2; portset.family= AF_UNIX;
                  }
                ;

type            : /* EMPTY */ { portset.type= SOCK_STREAM; }
                | TYPE TCP    { portset.type= SOCK_STREAM;  
		                Run.servicessl= TRUE; }
                | TYPE TCPSSL sslversion certmd5  { portset.type= SOCK_STREAM;
		                                    portset.ssl= TRUE; }
                | TYPE UDP    { portset.type= SOCK_DGRAM; }
                ;

certmd5         : /* EMPTY */    { portset.certmd5= NULL; }
                | CERTMD5 STRING { portset.certmd5= $2; }
                | EXPECT CERTMD5 STRING { portset.certmd5= $3; }
                ;

sslversion      : /* EMPTY */  { portset.sslversion= SSL_VERSION_AUTO; }
                | SSLV2        { portset.sslversion= SSL_VERSION_SSLV2; }
                | SSLV3        { portset.sslversion= SSL_VERSION_SSLV3; }
                | TLSV1        { portset.sslversion= SSL_VERSION_TLS; }
                | SSLAUTO      { portset.sslversion= SSL_VERSION_AUTO; }
                ;

protocol        : /* EMPTY */  { portset.protocol= addprotocol(P_DEFAULT); }
                | PROTOCOL DEFAULT { portset.protocol= addprotocol(P_DEFAULT); }
                | PROTOCOL HTTP request {portset.protocol= addprotocol(P_HTTP);}
                | PROTOCOL FTP { portset.protocol= addprotocol(P_FTP); }
                | PROTOCOL SMTP { portset.protocol= addprotocol(P_SMTP); }
                | PROTOCOL POP { portset.protocol= addprotocol(P_POP); }
                | PROTOCOL IMAP { portset.protocol= addprotocol(P_IMAP); }
                | PROTOCOL NNTP { portset.protocol= addprotocol(P_NNTP); }
                | PROTOCOL SSH  { portset.protocol= addprotocol(P_SSH); }
                | PROTOCOL DWP  { portset.protocol= addprotocol(P_DWP); }
                | PROTOCOL LDAP2 { portset.protocol= addprotocol(P_LDAP2); }
                | PROTOCOL LDAP3 { portset.protocol= addprotocol(P_LDAP3); }
                ;     
		   
request         : /* EMPTY */
                | REQUEST PATH  { portset.request= url_encode($2); free($2); }
                ;

timeout         : TIMEOUT NUMBER NUMBER {
                   check_timeout($2, $3);
		   current->def_timeout= TRUE;
		   current->to_start= $2;
		   current->to_cycle= $3;
                 }
                ;

alert           : alertmail '{' alertoptionlist '}' formatlist {
                   addmail($<string>1, &mailset, alertset);
                 }
                | alertmail formatlist {
                   addmail($<string>1, &mailset, alertset);
		 }
                ;

alertmail       : ALERT MAILADDR { $<string>$= $2; }
                | ALERT STRING { $<string>$= append_hostname($2); }
		;

alertoptionlist : alertoption
                | alertoptionlist alertoption
                ;

alertoption     : TIMEOUT   { alertset.alert_on_timeout= TRUE; }
                | RESTART   { alertset.alert_on_restart= TRUE; }
                | CHECKSUM  { alertset.alert_on_checksum= TRUE; }
                | RESOURCE  { alertset.alert_on_resource= TRUE; }
                | STOP      { alertset.alert_on_stop= TRUE; }
                | TIMESTAMP { alertset.alert_on_timestamp= TRUE; }
                ;

formatlist      : /* EMPTY */
	        | MAILFORMAT '{' formatoptionlist '}'
                ;

formatoptionlist: formatoption
                | formatoptionlist formatoption
                ;

formatoption    : MAILFROM { mailset.from= $1; }
                | MAILSUBJECT { mailset.subject= $1; }
                | MAILBODY { mailset.message= $1; }
                ;

every           : EVERY NUMBER {
                   check_every($2);
		   current->def_every= TRUE;
		   current->every= $2;
                 }
                ;

checksum        : CHECKSUM checksumlist { current->def_checksum= TRUE; }
                ;

checksumlist    : program
                | checksumlist program 
                ;

program         : PATH { addchecksum($1, NULL); }
                | PATH EXPECT STRING { addchecksum($1, $3); }
                ;

autostart       : AUTOSTART {
                    YERROR("Error: 'autostart' is obsolete - "
			   "use the 'mode' statement instead");
                  }
                ;

mode            : MODE ACTIVE  { current->mode= MODE_ACTIVE; }
                | MODE PASSIVE { current->mode= MODE_PASSIVE; }
                | MODE MANUAL  { current->mode= MODE_MANUAL;
		    current->do_validate= FALSE;
		  }
		;

group		: GROUP STRING { current->group= $2; }
		;

depend          : DEPENDS dependlist
                ;

dependlist      : dependant
                | dependlist dependant
                ;
 
dependant       : SERVICENAME { adddependant($<string>1); }
                ;

resource        : /* Old syntax */
                  CPUUSAGE operator value { 
		    YERROR("Error: 'CPUUSAGE' is obsolete -"
			   " use the 'CPU' statement instead");
                  }
                | MEMUSAGE operator value { 
		    YERROR("Error: 'MEMUSAGE' is obsolete -"
			   " use the 'MEMORY' statement instead");
		  }
                | MEMKBYTE operator value { 
		    YERROR("Error: 'MEMKBYTE' is obsolete -"
			   " use the 'MEMORY' statement instead");
		  }
                /* New syntax */
                | MEMORY operator value unit {
                    resourceset.resource_id= RESOURCE_ID_MEM_KBYTE;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($<real>3 * 
					      ($<number>4 / 1024.0)); 
                  }
                | MEMORY operator PERCENT {
                    resourceset.resource_id= RESOURCE_ID_MEM_PERCENT;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($<number>3 * 10); 
                  }
                | TOTALMEMORY operator value unit {
                    resourceset.resource_id= RESOURCE_ID_TOTAL_MEM_KBYTE;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($<real>3 * 
					      ($<number>4 / 1024.0)); 
                  }
                | TOTALMEMORY operator PERCENT  {
                    resourceset.resource_id= RESOURCE_ID_TOTAL_MEM_PERCENT;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($<real>3 * 10); 
                  }
                | CPU operator PERCENT {
                    resourceset.resource_id= RESOURCE_ID_CPU_PERCENT;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($<real>3 * 10); 
                  }
                /* Remaining syntax */
                | CHILDREN operator NUMBER { 
                    resourceset.resource_id= RESOURCE_ID_CHILDREN;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) $3; 
		  }
                | resourceload operator value { 
                    resourceset.resource_id= $<number>1;
                    resourceset.operator= $<number>2;
		    resourceset.limit= (int) ($<real>3 * 10.0); 
		  }
                ;

value           : REAL { $<real>$ = $1; }
                | NUMBER { $<real>$ = (float) $1; }
                ;

resourceload    : LOADAVG1  { $<number>$= RESOURCE_ID_LOAD1; }
                | LOADAVG5  { $<number>$= RESOURCE_ID_LOAD5; }
                | LOADAVG15 { $<number>$= RESOURCE_ID_LOAD15; }
                ;

resourcecycle   : /* EMPTY */
                | NUMBER { resourceset.max_cycle= $1; }
                ;

timestamp       : TIMESTAMP PATH operator NUMBER time action {
                    timestampset.pathname= $2;
		    timestampset.operator= $<number>3;
		    timestampset.time= ($4 * $<number>5);
		    timestampset.action= $<number>6;
                  }
                ;

operator        : /* EMPTY */ { $<number>$= OPERATOR_EQUAL; }
                | GREATER     { $<number>$= OPERATOR_GREATER; }
                | LESS        { $<number>$= OPERATOR_LESS; }
                | EQUAL       { $<number>$= OPERATOR_EQUAL; }
                | NOTEQUAL    { $<number>$= OPERATOR_NOTEQUAL; }
                ;

time            : /* EMPTY */ { $<number>$= TIME_SECOND; }
                | SECOND      { $<number>$= TIME_SECOND; }
                | MINUTE      { $<number>$= TIME_MINUTE; }
                | HOUR        { $<number>$= TIME_HOUR; }
                | DAY         { $<number>$= TIME_DAY; }
                ;

action          : ALERT       { $<number>$= ACTION_ALERT; }
                | RESTART     { $<number>$= ACTION_RESTART; }
                | STOP        { $<number>$= ACTION_STOP; }
                ;

inode           : INODE operator NUMBER action {
                    deviceset.resource= RESOURCE_ID_INODE;
                    deviceset.operator= $<number>2;
                    deviceset.limit_absolute= $3;
                    deviceset.action= $<number>4;
                  }
                | INODE operator PERCENT action {
                    deviceset.resource= RESOURCE_ID_INODE;
                    deviceset.operator= $<number>2;
                    deviceset.limit_percent= (int) $<real>3;
                    deviceset.action= $<number>4;
                  }
                ;

space           : SPACE operator value unit action {

                    DeviceInfo_T di= NEW(di);

                    get_fsusage(current->path, di);

                    deviceset.resource= RESOURCE_ID_SPACE;
                    deviceset.operator= $<number>2;
                    deviceset.limit_absolute= (int)((float)$<real>3 / 
						    (float)di->f_bsize * 
						    (float)$<number>4 );
                    deviceset.action= $<number>5;

                    free(di);

                  }
                | SPACE operator PERCENT action {
                    deviceset.resource= RESOURCE_ID_SPACE;
                    deviceset.operator= $<number>2;
                    deviceset.limit_percent= (int) $<real>3;
                    deviceset.action= $<number>4;
                  }
                ;

unit            : BYTE     { $<number>$= UNIT_BYTE; }
                | KILOBYTE { $<number>$= UNIT_KILOBYTE; }
                | MEGABYTE { $<number>$= UNIT_MEGABYTE; }
                | GIGABYTE { $<number>$= UNIT_GIGABYTE; }
                ;

perm            : PERM NUMBER { current->perm=check_perm($2); }
                ;

%%


/* -------------------------------------------------------- Parser interface */


/**
 * Syntactic error routine
 */
void yyerror(const char *msg) {

  ASSERT(msg);

  log("%s: %s '%s' at line %d\n", prog, msg, yytext, lineno);
  cfg_errflag++;

}


/*
 * The Parser hook - start parsing the control file
 * Returns TRUE if parsing succeeded, otherwise FALSE
 */
int parse(char *controlfile) {

  ASSERT(controlfile);

  servicelist= tail= current= NULL;
  
  if(! check_rcfile(controlfile)) {
    
    return (FALSE);
    
  }
  
  if ((yyin = fopen(controlfile,"r")) == (FILE *)NULL) {
    
    log("%s: Error opening the control file '%s' -- %s\n",
	prog, controlfile, STRERROR);
    return(FALSE);
    
  }

  /* Creation of the global service list is synchronized  */
  LOCK(Run.mutex)
      
      initialize();
      yyparse();
      fclose(yyin);

      /* If defined - add the last service to the service list */
      if(current) {
	
	createinfo();
	addservice(current);
	
      }
      
      /* Check the sanity of any dependency graph */
      check_depend();
  END_LOCK;
    
  return(cfg_errflag == 0);
  
}


/* ----------------------------------------------------------------- Private */


/**
 * Initialize objects. Placeholder for setting values that can change
 * between cycles (changed config file) and not easily supported by
 * the grammar implementation.
 */
static void initialize() {

  destroy_hosts_allow(); 
  Run.dolog= FALSE;
  Run.dohttpd= FALSE;
  Run.Auth.defined= FALSE;
  Run.servicessl= FALSE;
  Run.clientssl= FALSE;
  depend_list= NULL;
  reset_runmail();
  if(Run.bind_addr) {
    free(Run.bind_addr);
    Run.bind_addr= NULL;
  }
  Run.httpdssl= FALSE;
  
}


/*
 * Create a new service object and add any current objects to the
 * service list.
 */
static void createservice(int task, char *name, char *path) {

  ASSERT(name);
  ASSERT(path);

  if(current) {

    createinfo();
    addservice(current);
    
  }
  else {
    
    current= NEW(current);
    
  }

  /* Reset the current object */
  memset(current, 0, sizeof(*current));

  /* Set default values */
  current->do_validate= TRUE;
  current->mode= MODE_ACTIVE;

  current->task= task;
  current->name= name;
  current->path= path;
  current->group= NULL;
  current->dependantlist= NULL;
  current->perm= -1;

  pthread_mutex_init(&current->mutex, NULL);

}


/*
 * Add a service object to the servicelist
 */
static void addservice(Service_T s) {
  
  Service_T n;

  ASSERT(s);
 
  n= NEW(n);
  memcpy(n, s, sizeof(*s));

  /* Add the service to the end of the service list */
  if(tail != NULL) {
    
    tail->next= n;
    
  } else {
    
    servicelist= n;
    
  }
  
  tail= n;

}


/* 
 * Add a dependant entry to the current service dependant list
 *
 */
static void adddependant(char *dependant) {

  Dependant_T d= NEW(d);

  ASSERT(dependant);

  if (current->dependantlist != NULL) {
    d->next= current->dependantlist;
  }

  d->dependant=dependant;
  current->dependantlist= d;

}


/*
 * Add a filename and its associated md5 checksum
 * to the current service checksumlist
 */
static void addchecksum(char *filename, char *sum) {

  Checksum_T c= NEW(c);

  ASSERT(filename);

  c->file= filename;
  
  if(sum) {
    
    c->md5= sum;

    
  } else if(! set_md5sum(&c->md5, filename)) {
    
    log("%s: Cannot compute a checksum for %s at line %d\n",
	prog, filename, lineno-1);
    cfg_errflag++;

  }

  
  if (c->md5 != NULL) {
   
    cleanup_md5_string(c->md5);

  }

  c->next= current->checksumlist;
  current->checksumlist= c;

}


/*
 * Add the given mailaddress with the apropriate alert notification
 * mask and mail attributes to the current service's mailinglist.
 */
static void addmail(char *mailto, Mail_T f, Alert_T a) {

  Mail_T m= NEW(m);

  ASSERT(mailto);

  if(a.alert_on_timeout  || a.alert_on_restart  ||
     a.alert_on_checksum || a.alert_on_resource ||
     a.alert_on_stop     || a.alert_on_timestamp) {
    
    memcpy(&m->alerts, &a, sizeof(Alert_T));
    
  } else {

    /* The following turns on all alerts (i.e. to TRUE). NB! remember
     * to update the initializer below if new members are added to the
     * Alert_T object */
    Alert_T tmp= {1, 1, 1, 1, 1, 1};
    memcpy(&m->alerts, &tmp, sizeof(Alert_T));
    
  }
  
  m->to= mailto;
  m->from= f->from;
  m->subject= f->subject;
  m->message= f->message;
  
  m->next= current->maillist;
  current->maillist= m;

  reset_mailset();

}


/*
 * Add the given portset to the current service's portlist
 */
static void addport(struct PortSet *pp) {

  char address[STRLEN];
  Port_T p= NEW(p);

  ASSERT(pp);

  p->socket= pp->socket;
  p->hostname= pp->hostname;
  p->port= pp->port;
  p->type= pp->type;
  p->family= pp->family;
  p->request= pp->request;
  p->protocol= pp->protocol;
  p->pathname= pp->pathname;
  p->certmd5= pp->certmd5;

  if (p->certmd5 != NULL) {
    
    cleanup_md5_string(p->certmd5);

  }

  if(pp->family == AF_INET) {
    snprintf(address, STRLEN, "INET[%s:%d]", pp->hostname, pp->port);
  } else if(pp->family == AF_UNIX) {
    snprintf(address, STRLEN, "UNIX[%s]", pp->pathname);
  }

  if(pp->ssl == TRUE) {

    if(!have_ssl()) {

      log("%s: The ssl check cannot be activated at line %d, " 
	  "SSL is not support.\n",
	  prog,lineno-1);
      cfg_errflag++;

    } else {

      p->ssl= new_ssl_connection(NULL, pp->sslversion);

    }

  } else {

    p->ssl= NULL;

  }

  p->address= xstrdup(address);
  
  p->next= current->portlist;
  current->portlist= p;
  
  reset_portset();

}


/*
 * Add a new resource object to the current service resource list
 */
static void addresource(struct ResourceSet *rr) {

  Resource_T r= NEW(r);

  ASSERT(rr);

  if(! Run.doprocess) {

    log("%s: Cannot activate service check at line %d\n"
	"\t(The process status engine was disabled. On certain"
	" systems you must\n\trun monit as root to utilize this"
	" feature)\n", prog, lineno-1);
    cfg_errflag++;
    
  }

  r->resource_id= rr->resource_id;
  r->limit= rr->limit;
  r->cycle= 0;
  r->max_cycle= rr->max_cycle;
  r->action= rr->action;
  r->operator= rr->operator;
  r->next= current->resourcelist;

  if(r->max_cycle < 1) {

    log("%s: The number of cycles must be greater then 0 in line %d\n",
	prog, lineno-1);
    cfg_errflag++;
    
  }

  current->resourcelist= r;
  reset_resourceset();

}


/*
 * Add a new file object to the current service timestamp list
 */
static void addtimestamp(struct TimestampSet *ts) {

  Timestamp_T t= NEW(t);

  ASSERT(ts);

  t->pathname= ts->pathname;
  t->operator= ts->operator;
  t->time= ts->time;
  t->action= ts->action;

  t->next= current->timestamplist;
  current->timestamplist= t;

  reset_timestampset();

}


/*
 * Add a new device to the current tasks device list
 */
static void adddevice(struct DeviceSet *ds) {

  Device_T d= NEW(d);

  ASSERT(ds);

  d->resource= ds->resource;
  d->operator= ds->operator;
  d->limit_absolute= ds->limit_absolute;
  d->limit_percent= ds->limit_percent;
  d->action= ds->action;

  d->next= current->devicelist;
  current->devicelist= d;

  reset_deviceset();

}


/*
 * Adds runtime info to current service
 */
static void createinfo() {

  DeviceInfo_T di= NEW(di);
  ProcInfo_T pi= NEW(pi);

  current->devinfo=di;
  current->procinfo=pi;
  
}


/*
 * Return a protocol object for the given protocol
 */
static void *addprotocol(int proto) {

  switch (proto) {
    
    case P_HTTP:  return create_http();
    case P_FTP:   return create_ftp();
    case P_SMTP:  return create_smtp();
    case P_POP:   return create_pop();
    case P_IMAP:  return create_imap();
    case P_NNTP:  return create_nntp();
    case P_SSH:   return create_ssh();
    case P_DWP:   return create_dwp();
    case P_LDAP2: return create_ldap2();
    case P_LDAP3: return create_ldap3();
	
  }

  return create_default();
   
}


/*
 * Add the current command object to the current service object
 */
static void addcommand(int what) {

  switch(what) {
    
    case START: current->start= command; break;
    case STOP:  current->stop= command; break;
	
  }
  
  command= NULL;
  
}

  
/*
 * Add a new argument to the argument array
 */
static void addargument(char *argument) {

  ASSERT(argument);

  if(! command) {
    
    command= NEW(command);

  }
  
  command->arg[command->length++]= argument;
  command->arg[command->length]= NULL;
  
  if(command->length >= ARGMAX) {
    
    log("%s: Exceeded maximum number of program arguments at line %d\n",
	prog, lineno-1);
    cfg_errflag++;
    
  }
  
}


/*
 * Add a new user id to the current command object. If the parameter,
 * user is NULL the uid parameter is used for looking up the user id
 * on the system, otherwise the user parameter is used.
 */
static void adduid(char *user, uid_t uid) {

  if(!getuid()) {

#ifdef HAVE_GETPWNAM_R
    char buf[1024];
    struct passwd pwd;
#endif
    struct passwd *r;

    if(user) {
      
#ifdef HAVE_GETPWNAM_R
      if((0 != getpwnam_r(user, &pwd, buf, 1024, &r)) || !r) {
#else
	r = getpwnam(user);
	if ( r == NULL ) {
#endif
	
       log("%s: Requested user '%s' not found on the system. Error at line %d\n",
	   prog, user, lineno);
       cfg_errflag++;
      
      }
      
    } else {
      
#ifdef HAVE_GETPWUID_R
      if((0 != getpwuid_r(uid, &pwd, buf, 1024, &r)) || !r) {
#else
      r = getpwuid(uid);
      if ( r == NULL ) {
#endif     
	
       log("%s: Requested uid '%d' not found on the system. Error at line %d\n",
	   prog, uid, lineno);
       cfg_errflag++;
      
      }
      
    }
  
    if(r) {
      
      command->has_uid= TRUE;
      command->uid= r->pw_uid;
      
      }
    
  } else {
    
    log("%s: Ignoring uid statement at line %d, not running as root.\n",
	prog, lineno);
    
  }
  
}


/*
 * Add a new group id to the current command object. If the parameter
 * group is NULL the gid parameter is used for looking up the group id
 * on the system, otherwise the group parameter is used.
 */
static void addgid(char *group, gid_t gid) {

  if(!getuid()) {

#ifdef HAVE_GETGRNAM_R    
    char buf[1024];
    struct group gwd;
#endif
    struct group *r;

    if(group) {
      
#ifdef HAVE_GETGRNAM_R
      if((0 != getgrnam_r(group, &gwd, buf, 1024, &r)) || !r) {
#else
      r = getgrnam(group);
      if ( r == NULL ) {
#endif	
	
      log("%s: Requested group '%s' not found on the system. Error at line %d\n",
	  prog, group, lineno);
      cfg_errflag++;
      
      }
      
    } else {
	
#ifdef HAVE_GETGRGID_R
      if((0 != getgrgid_r(gid, &gwd, buf, 1024, &r)) || !r) {
#else
      r = getgrgid(gid);
      if ( r == NULL ) {
#endif
	
      log("%s: Requested gid '%d' not found on the system. Error at line %d\n",
	  prog, gid, lineno);
      cfg_errflag++;
      
      }

    }
    
    if(r) {
      
      command->has_gid= TRUE;
      command->gid= r->gr_gid;
      
    }
    
  } else {
    
    log("%s: Ignoring gid statement at line %d, not running as root.\n",
	prog, lineno);
    
  }

}


/*
 * Reset the logfile if changed
 */
static void setlogfile(char *logfile) {

  if(Run.logfile) {
    
    if(IS(Run.logfile, logfile)) {
      
      return;
      
    } else {
      
      free(Run.logfile);
      
    }
  }
  
  Run.logfile= logfile;
 
}


/*
 * Reset the pidfile if changed
 */
static void setpidfile(char *pidfile) {

  if(Run.pidfile) {
    
    if(IS(Run.pidfile, pidfile)) {
      
      return;
      
    } else {
      
      free(Run.pidfile);
      
    }
  }
  
  Run.pidfile= pidfile;
 
}


/*
 * Set mailserver if changed
 */
static void setmailserver(char *mailserver) {

  ASSERT(mailserver);

  if(Run.mailserver) {
    
    if(IS(Run.mailserver, mailserver)) {
      
      return;
      
    } else {
      
      free(Run.mailserver);
      
    }
  }
  
  Run.mailserver= mailserver;

}


/*
 * Set Basic Authentication username
 */
static void setuname(char *uname) {

  ASSERT(uname);

  if(Run.Auth.uname) {
    
    if(IS(Run.Auth.uname, uname)) {
      
      return;
      
    } else {
      
      free(Run.Auth.uname);
      
    }
  }
  
  Run.Auth.uname= uname;

}

 
/*
 * Set Basic Authentication password
 */
static void setpasswd(char *passwd) {

  ASSERT(passwd);

  if(Run.Auth.passwd) {
    
    if(IS(Run.Auth.passwd, passwd)) {
      
      return;
      
    } else {
      
      free(Run.Auth.passwd);
      
    }
  }
  
  Run.Auth.passwd= passwd;

} 


/*
 * Set the syslog and the facilities to be used
 */
static void setsyslog(char *facility) {

  if (!Run.logfile || ihp.logfile) {

    ihp.logfile= TRUE;
    setlogfile(xstrdup("syslog"));
    Run.use_syslog= TRUE;
    Run.dolog=TRUE;

  }

  if(facility) {
    if(IS(facility,"log_local0")) {
      Run.facility = LOG_LOCAL0;
    } else if(IS(facility, "log_local1")) {
      Run.facility = LOG_LOCAL1;
    } else if(IS(facility, "log_local2")) {
      Run.facility = LOG_LOCAL2;
    } else if(IS(facility, "log_local3")) {
      Run.facility = LOG_LOCAL3;
    } else if(IS(facility, "log_local4")) {
      Run.facility = LOG_LOCAL4;
    } else if(IS(facility, "log_local5")) {
      Run.facility = LOG_LOCAL5;
    } else if(IS(facility, "log_local6")) {
      Run.facility = LOG_LOCAL6;
    } else if(IS(facility, "log_local7")) {
      Run.facility = LOG_LOCAL7;
    } else if(IS(facility, "log_daemon")) {
      Run.facility = LOG_DAEMON;
    } else {
      log("Invalid syslog facility '%s' used at line %d\n", facility, lineno);
      cfg_errflag++;
    }
  } else {
    Run.facility= LOG_USER;
  }
  
}


/*
 * Reset the current mailset and alertset for reuse
 */
static void reset_mailset() {

  memset(&mailset, 0, sizeof(struct mymail));
  memset(&alertset, 0, sizeof(struct myalert));

}


/*
 * Reset the Port set to default values
 */
static void reset_portset() {

  portset.socket= -1;
  portset.hostname= NULL;
  portset.port= 0;
  portset.type= SOCK_STREAM;
  portset.family= AF_INET;
  portset.ssl= FALSE;
  portset.sslversion= SSL_VERSION_AUTO;
  portset.request= NULL;
  portset.protocol= NULL;
  portset.pathname= NULL;

}


/*
 * Reset the Proc set to default values
 */
static void reset_resourceset() {

  resourceset.resource_id= 0;
  resourceset.limit= 0;
  resourceset.max_cycle= 1;
  resourceset.action= ACTION_ALERT;
  resourceset.operator= OPERATOR_EQUAL;

}


/*
 * Reset and release the Run mail-format sub-structure
 */
static void reset_runmail() {

  free(Run.MailFormat.from);
  Run.MailFormat.from= NULL;
  free(Run.MailFormat.subject);
  Run.MailFormat.subject= NULL;
  free(Run.MailFormat.message);
  Run.MailFormat.message= NULL;

}


/*
 * Reset the Timestamp set to default values
 */
static void reset_timestampset() {

  timestampset.pathname= NULL;
  timestampset.operator= OPERATOR_EQUAL;
  timestampset.time= 0;
  timestampset.action= ACTION_ALERT;

}


/*
 * Reset the Device set to default values
 */
static void reset_deviceset() {

  deviceset.resource= 0;
  deviceset.operator= OPERATOR_EQUAL;
  deviceset.limit_absolute= -1;
  deviceset.limit_percent= -1;
  deviceset.action= ACTION_ALERT;

}


/* ---------------------------------------------------------------- Checkers */


/*
 * Check for unique service name
 */
static void check_name(char *name) {

  ASSERT(name);

  if(exist_service(name)) {
    
    log("%s: Name conflict at line %d - check name '%s' already defined\n",
	prog, lineno, name);
    
    cfg_errflag++;
    
  }
  
}


/*
 * Permission statement semantic check
 */
static int check_perm(int perm) {

  long result;
  char *status;
  char buf[STRLEN];

  snprintf(buf, STRLEN, "%d", perm);

  result= strtol(buf, &status, 8);

  if( *status != '\0' || result < 0 || result > 07777 ) {
    
    yyerror("An permission statement must have an octal value between 0 "
	    "and 07777.");
    
  }

  return result;
  
}


/*
 * Timeout statement semantic check
 */
static void check_timeout(int s, int c) {

  if(s > c) {
    
    yyerror("The number of restarts must be less than poll cycles");
    
  }

  if(s <= 0 || c <= 0) {
    
    yyerror("Zero or negative values not allowed in a timeout statement");
    
  }
  
}


/*
 * Every statement semantic check
 */
static void check_every(int every) {

  if(every <= 1) {
    
    yyerror("An every statement must have a value greater than 1.");
    
  }
  
}


/*
 * Check hostname 
 */
static void check_hostname(char *hostname) {

  ASSERT(hostname);

  if(!check_host(hostname)) {
    
    yyerror("Hostname did not resolve");
    
  }
 
}


/*
 * Check the dependency graph for errors and if dependencies are
 * present reshuffle the service list in a depending order.
 */
static void check_depend() {

  Service_T s;
  int has_depend= FALSE;
  
  for(s= servicelist; s; s= s->next) {
    if(s->visited)
	continue;
    validate_depend(s, &has_depend);
    reset_depend();
  }
  
  if(has_depend) {
    
    Service_T d;
    
    for(s= servicelist; s; s= s->next) {
      if(s->visited)
	  continue;
      order_depend(s);
    }

    ASSERT(depend_list);
    servicelist= depend_list;
    
    for(d= depend_list; d; d= d->next_depend)
	d->next= d->next_depend;
    
  }

  reset_depend();
  
}



/* -------------------------------------------------------------------- Misc */


/*
 * Create a real email address from a username
 */
static char *append_hostname(char *addr) {

  char buf[STRLEN];
  char * fulladdr;

  ASSERT(addr);

  if (gethostname(buf, STRLEN) != 0) {
     yyerror("Cannot find hostname for mailaddress");
  }

  fulladdr = xresize(addr, strlen(buf)+1);
  strcat(fulladdr, "@");
  strcat(fulladdr, buf);

  return fulladdr;

}


/*
 * Cleans up an md5 string, tolower and remove byte sperators
 */
static void cleanup_md5_string(char *md5string) {

  int i= 0, j= 0;

  ASSERT(md5string);

  while (md5string[i] != '\0') {

    if (isxdigit((int) md5string[i])) {
      
      md5string[j]=tolower(md5string[i]);
      j++;

    } 

    i++;

  }
  
  md5string[j]='\0';

  return;
 
}


/*
 * Search for any errors in the service dependency graph
 */
static void validate_depend(Service_T s, int *has_depend) {

  ASSERT(s);

  if(s->visited)
      return;
  
  if(s->dependantlist) {
    
    Dependant_T d;
    
    for(d= s->dependantlist; d; d= d->next) {
      
      Service_T dp= get_service(d->dependant);
      
      if(!dp) {
	log("%s: Error, depend service '%s' is not defined in the "
	    "control file\n", prog, d->dependant);
	exit(1);
      }
      
      if(dp->depend_visited) {
	log("%s: Error, found a depend loop in the control file involving "
	    "the service '%s'\n", prog, s->name);
	exit(1);
      }
      
      *has_depend= TRUE;
      dp->depend_visited= TRUE;
      validate_depend(dp, has_depend);
      
    }
  }
  
  s->visited= TRUE;

}


/*
 * Order the service list with the most "depending" service last and
 * the least first.
 */
static void order_depend(Service_T s) {

  ASSERT(s);
  
  if(s->visited)
      return;

  s->visited= TRUE;

  if(s->dependantlist) {
    
    Dependant_T d;
    
    for(d= s->dependantlist; d; d= d->next) {
      
      Service_T dp= get_service(d->dependant);
      
      order_depend(dp);
      
    }
  }

  s->next_depend= depend_list;
  depend_list= s;

}
