/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include "alert.h"
#include "monitor.h"
#include "engine.h"


/**
 *  Function for spawning of a process. This function fork's twice to
 *  avoid creating any zombie processes. Inspired by code from
 *  W. Richard Stevens book, APUE.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *
 *  @version \$Id: spawn.c,v 1.23 2003/06/24 02:49:09 hauk Exp $
 *
 *  @file
 */


/* -------------------------------------------------------------- Prototypes */


static void exec_alert_mail(Service_T s, Command_T c);
static void set_monit_environment(Service_T s, Command_T c);


/* ------------------------------------------------------------------ Public */


/**
 * Execute the given command. If the execution fails, an alert message
 * is sent to the email addresses found in the service object that has
 * registred interest for restart alerts.
 * @param P A Service object
 * @param C A Command object
 */
void  spawn(Service_T S, Command_T C) {
  
  pid_t pid;
  sigset_t mask, save;

  ASSERT(S);
  ASSERT(C);

  /*
   * Block SIGCHLD
   */
  sigemptyset(&mask);
  sigaddset(&mask, SIGCHLD);
  pthread_sigmask(SIG_BLOCK, &mask, &save);
  
  if((pid= fork()) < 0) {
    
    log("Cannot fork of a new process\n");  
    exit(1);
    
  }  else if(pid == 0) {
    
    if((pid= fork()) < 0) {
      
      log("Cannot fork of a new process\n");  
      exit(1);
      
    }
    
    else if(pid > 0) {
      
      _exit(0);
      
    } else  {

      
      /*
       * Reset all signals, so the child process is
       * *not* created with any inherited SIG_BLOCK
       */
      sigemptyset(&mask);
      pthread_sigmask(SIG_SETMASK, &mask, NULL);
      
      /* Unregister Signal handlers */
      signal(SIGTERM, SIG_DFL);
      signal(SIGUSR1, SIG_DFL);
      signal(SIGPIPE, SIG_DFL);

      /*
       * Reset to the original umask so programs will inherit the
       * same file creation mask monit was started with
       */
      umask(Run.umask);

      /*
       * Redirect the standard file descriptors
       */
      redirect_stdfd();

      /*
       * Switch uid and gid if requested
       */
      if(C->has_gid) {
	if(0 != setgid(C->gid)) {
	  log("Failed to change gid to '%d' for '%s'\n",
	      C->gid, C->arg[0]);
	}
      }
      if(C->has_uid) {
	if(0 != setuid(C->uid)) {
	  log("Failed to change uid to '%d' for '%s'\n",
	      C->uid, C->arg[0]);
	}
      }

      /*
       * Set the MONIT_xxx environment variables
       */
      set_monit_environment(S, C);
      
      /*
       * The exec statement done by the second child
       */
      (void) execv(C->arg[0], C->arg);
      exec_alert_mail(S, C);
      log("Could not execute %s\n", C->arg[0]);
      _exit(1);
      
    }
  }

  /* Wait for first child - aka second parent, to exit */
  if(waitpid(pid, NULL, 0) != pid) {
    
      log("Waitpid error\n");

  }

  /*
   * Restore the signal mask
   */
  pthread_sigmask(SIG_SETMASK, &save, NULL);
  
  /*
   * We do not need to wait for the second child since we forked twice,
   * the init system-process will wait for it. So we just return
   */
  return ;
  
} 


/* ----------------------------------------------------------------- Private */


/*
 * Send an exec failed alert mail to those mail objects that has
 * registred interest for restart notification.
 */
static void exec_alert_mail(Service_T s, Command_T c) {

  Mail_T a= s->maillist;

  for(; a; a= a->next)
      if(a->alerts.alert_on_restart) {
	struct mymail m;
	char message[STRLEN];
	
	m.to= a->to;
	m.from= "monit@localhost";
	m.subject= "monit alert -- exec error";
	snprintf(message, STRLEN,
		 "Could not execute program '%s' for service %s\n\n"
		 "Your faithful employee,\nmonit",
		 c->arg[0]?c->arg[0]:"null", s->name);
	m.message= message;
	sendmail(&m);
      }
  
}


/*
 * Setup the environment with special MONIT_xxx variables. The program
 * executed may use such variable for various purposes.
 */
static void set_monit_environment(Service_T s, Command_T c) {

  int i;
  char buf[10][STRLEN];

  snprintf(buf[0], STRLEN, "MONIT_PROGRAM_PID=%d", is_process_running(s));
  snprintf(buf[1], STRLEN, "MONIT_PROGRAM=%s", c->arg[0]);
  snprintf(buf[2], STRLEN, "MONIT_DATE=%s", get_RFC822date(NULL));
  snprintf(buf[3], STRLEN, "MONIT_HOST=%s", Run.localhostname);
  snprintf(buf[4], STRLEN, "MONIT_PROGRAM_MEMORY=%ld", s->procinfo->mem_kbyte);
  snprintf(buf[5], STRLEN, "MONIT_PROGRAM_CHILDREN=%d", s->procinfo->children);
  snprintf(buf[6], STRLEN, "MONIT_PROGRAM_CPU_PERCENT=%d",
	   s->procinfo->cpu_percent);
  
  buf[7][0]= 0;

  for(i= 0; *buf[i]; i++)
      putenv(buf[i]);

}
