/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <sys/socket.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "net.h"
#include "socket.h"
#include "monitor.h"
#include "monit_process.h"



/* Private Prototypes */
static void local_status(Service_T);
static int remote_status(Service_T);

/**
 *  Print the status of services in the service list.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Christian Hopp, <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: status.c,v 1.29 2003/06/17 17:54:15 hauk Exp $
 *
 *  @file
 */

/* ------------------------------------------------------------------ Public */


/**
 * Show all services in the service list
 */
void status() {

  Service_T s;
  int remote= TRUE;
  char *uptime= get_process_uptime(Run.pidfile);

  fprintf(stdout, "monit daemon uptime: %s\n", uptime);
  free(uptime);
  
  for(s= servicelist; s; s= s->next) {

    if(remote) {

      remote= remote_status(s);
      
    } else {
      
      local_status(s);
      
    }
    
  }

}

/**
 * Show all services in the group
 * @param G  group name
 */
void status_group(char *G) {

  Service_T s;
  int remote= TRUE;

  ASSERT(G);
  
  for(s= servicelist; s; s= s->next) {

    if(IS(s->group, G)) {
      
      if(remote) {
	
	remote= remote_status(s);

      } else {
	
	local_status(s);

      }

    }
    
  }
  
}


/* ----------------------------------------------------------------- Private */


/**
 * Print the service status for the given service
 * @param s  A Service_T object
 */
static void local_status(Service_T s) {
  
  pid_t  pid= -1;
  struct stat stat_buf;

  ASSERT(s);
  
  switch(s->task) {

  case TASK_PROCESS:
    if((pid= is_process_running(s))) {

      char *uptime= get_process_uptime(s->path);

      fprintf(stdout, "Process '%s' is running with pid [%d]\n\t"
	      "Uptime: %s\n\tMonitoring status: %s\n",
	      s->name, (int)pid, uptime, statusnames[s->do_validate]);
    
      free(uptime);
 
    } else {
    
      fprintf(stdout, "Process '%s' is not running\n",  s->name);
    
    }
    break;

  case TASK_DEVICE:
    if(!get_fsusage(s->path, s->devinfo)) {
      fprintf(stdout, "Device '%s' is not accessible\n\t"
        "Space: - Inodes: -\n\tMonitoring status: %s\n",
        s->name, statusnames[s->do_validate]);
    } else {

      if(s->devinfo->f_files > 0) {

        fprintf(stdout, "Device '%s' is accessible\n\t"
          "Space: %.1f%% [%.1f MB] Inodes: %.1f%% [%ld objects]\n\t"
          "Monitoring status: %s\n",
          s->name,
          (float) 100 * (s->devinfo->f_blocks - s->devinfo->f_blocksfree) / s->devinfo->f_blocks,
          (float) (s->devinfo->f_blocks - s->devinfo->f_blocksfree) / 1048576 * s->devinfo->f_bsize,
          (float) 100 * (s->devinfo->f_files - s->devinfo->f_filesfree) / s->devinfo->f_files,
          s->devinfo->f_files - s->devinfo->f_filesfree,
          statusnames[s->do_validate]);

      } else {

        fprintf(stdout, "Device '%s' is accessible\n\t"
		"Space: %.1f%% [%.1f MB]\n\tMonitoring status: %s\n",
		s->name,
          (float) 100 * (s->devinfo->f_blocks - s->devinfo->f_blocksfree) / s->devinfo->f_blocks,
          (float) (s->devinfo->f_blocks - s->devinfo->f_blocksfree) / 1048576 * s->devinfo->f_bsize,
          statusnames[s->do_validate]);

      }

    }
    break;

  case TASK_FILE:
    if( (stat(s->path, &stat_buf) != 0) || !S_ISREG(stat_buf.st_mode) ) {
      fprintf(stdout, "File '%s' doesn't exist\n\tMonitoring status: %s\n",
        s->name, statusnames[s->do_validate]);
    } else {
      fprintf(stdout,
        "File '%s' exist\n\t"
        "Size: %.3f MB UID: %d GID: %d Permission: %o\n\t"
        "Monitoring status: %s\n",
        s->name,
        (float)stat_buf.st_size/1048576, (int)stat_buf.st_uid,
        (int)stat_buf.st_gid, (int)(stat_buf.st_mode & 07777),
        statusnames[s->do_validate]);
    }
    break;

  case TASK_DIRECTORY:
    if( (stat(s->path, &stat_buf) != 0) || !S_ISDIR(stat_buf.st_mode) ) {
      fprintf(stdout, "Directory '%s' doesn't exist\n\tMonitoring status: %s\n",
        s->name, statusnames[s->do_validate]);
    } else {
      fprintf(stdout,
	      "Directory '%s' exist\n\t"
	      "UID: %d GID: %d Permission: %o\n\t"
	      "Monitoring status: %s\n",
	      s->name,
	      (int)stat_buf.st_uid, (int)stat_buf.st_gid,
	      (int)(stat_buf.st_mode & 07777),
	      statusnames[s->do_validate]);
    }
    break;

  default:
    break;

  }

}


/**
 * Get the service status for the given service from a monit server
 * @param p  A Service_T object
 * @return TRUE if the monit server could be connected 
 */
static int remote_status(Service_T s) {

  ASSERT(s);
  
  if(exist_daemon()) {
    
    /* If a monit daemon exist we request status information from the server */
    
    Socket_T sock= socket_new(Run.bind_addr?Run.bind_addr:"localhost",
			      Run.httpdport, SOCK_STREAM, Run.httpdssl);
    if(!sock) {

      fprintf(stdout,
	      "Cannot connect to monit server to get extended service data.\n");
      local_status(s);
      
      return FALSE;
            
    } else {

      int n;
      char buf[STRLEN];
      char *auth= get_basic_authentication_header();

      socket_print(sock, "GET /%s?action=status HTTP/1.0\r\n%s\r\n",
		   s->name, auth);
      free(auth);
      
      if(0>(n= socket_read(sock, buf, STRLEN))) {
	
	local_status(s);
	socket_free(&sock);
	
	return TRUE;
	
      }
      
      socket_free(&sock);
	

      /* If everything has gone well the returned string starts with
	 "Process ", "Device ", "File " or "Directory " */
      
      buf[n]= 0;
      if(starts_with(buf, "Process ") ||
         starts_with(buf, "Device ") ||
	 starts_with(buf, "File ") ||
	 starts_with(buf, "Directory ")) {
	
	fprintf(stdout, "%s", buf);
	
      } else {

	fprintf(stdout, "The monit server did not return a process, "
          "device, file or directory record.\n");
	local_status(s);
    
      }

      return TRUE;
    }
    
  } else {
    
    /* No monit daemon exist, just print local status information */
    
    fprintf(stdout, "Cannot connect to the monit server to get extended "
	    "process data.\n");
    local_status(s);
    
    return FALSE;
  }

}
