/*
 * Copyright (C), 2000-2003 by Contributors to the monit codebase. 
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <config.h>

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <setjmp.h>
#include <sys/socket.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "monitor.h"
#include "alert.h"
#include "net.h"
#include "monit_process.h"


/**
 *  This function contains the main process and check machinery for
 *  monit. The validate function check services in the service list
 *  to see if they are running and that they accepts connections if
 *  ports are defined. Not running services are restarted. This
 *  function also handles the various tests per service that are
 *  defined in the control file.
 *
 *  @author Jan-Henrik Haukeland, <hauk@tildeslash.com>
 *  @author Olivier Beyssac, <ob@r14.freenix.org> (check_skip)
 *  @author Martin Pala <martin.pala@iol.cz>
 *  @author Christian Hopp <chopp@iei.tu-clausthal.de>
 *
 *  @version \$Id: validate.c,v 1.45 2003/06/26 14:24:11 hauk Exp $
 *
 *  @file
 */


/* ------------------------------------------------------- Private variables */


static sigjmp_buf timeout;


/* -------------------------------------------------------------- Prototypes */


static int  check_skip(Service_T);
static void connection_timeout(int);
static int  check_timeout(Service_T);
static int  check_checksum(Service_T);
static int  do_not_validate(Service_T);
static int  compare_value(int, int, int);
static void do_stop(Service_T, char *, ...);
static void do_start(Service_T, char *, ...);
static void reset_resource_counter(Service_T);
static void do_restart(Service_T, char *, ...);
static int  check_process_state(Service_T, char *);
static int  check_device(Service_T, Device_T, char *);
static int  check_connection(Service_T, Port_T, char *);
static void do_validate(Service_T, ProcessTree_T *, int);
static int  check_resources(Service_T, Resource_T, char *);
static int  check_timestamp(Service_T, Timestamp_T, char *);
static void vlog(char * report, int n, Service_T s, char *m,...);


/* ---------------------------------------------------------------- Public */


/**
 * Walk the service-list and validate each service
 */
void validate() {

  Service_T s;
  int treesize= 0;
  ProcessTree_T *pt;

  if(Run.doprocess) {

    treesize = initprocesstree(&pt);

  }

  if(! update_loadavg()) {
    
    log("Update of loadavg has failed!\n");
    
  }
 
  for(s= servicelist; s; s= s->next) {
    if(s->visited)
	continue;
    do_validate(s, pt, treesize);
  }

  reset_depend();

  if (Run.doprocess) {

    delprocesstree(pt, treesize);

  }

}


/* --------------------------------------------------------------- Private */


/**
 * Validate the given service - If the service is not running, start
 * it. If the service listens on a port try to connect to the port and
 * test the stated protocol at the port. If any of the various tests
 * fails then do the specified action on the service.
 *
 * We block for the signals SIGTERM while in this function.
 */
static void do_validate(Service_T s, ProcessTree_T * pt, int treesize) {
  
  Port_T pp;
  Device_T td;
  Resource_T pr;
  Timestamp_T tl;
  pid_t  pid= -1;
  sigset_t ns,os;
  char report[STRLEN];
  struct stat stat_buf;

  ASSERT(s);
  ASSERT(s->path);
  
  /* Obtain service action mutex */
  LOCK(s->mutex)

  /* First, check for pre-conditions */
  if(do_not_validate(s)) {
    /* Release service action mutex */
    pthread_mutex_unlock(&s->mutex);
    return;
  }
  
  /* Set a SIGTERM block */
  sigemptyset(&ns);
  sigaddset(&ns, SIGTERM);
  pthread_sigmask(SIG_BLOCK, &ns, &os);


  /** Context specific task actions */
  switch(s->task) {

  case TASK_PROCESS:

    /* Test for running process */
    if(!(pid= is_process_running(s))) {

      /* Reset the proc info object to prevent false data in the first run */
      memset(s->procinfo, 0, sizeof *(s->procinfo));
    
      do_start(s, "Reason: Process is not running.");
      goto reinstall;
    
    } else {
    
        DEBUG("'%s' is running with pid %d\n", s->name, (int)pid);
    
    }

    if(Run.doprocess) {

      if(update_process_data(s, pt, treesize, pid)) {
  
        if(! check_process_state(s, report)) {

  	  alert_resource(s, "Reason: %s\n", report);
    
        } else {
      
	  DEBUG("'%s' check_process_state() passed.\n", s->name);
      
        }
    
        for(pr= s->resourcelist; pr; pr= pr->next) {
      
	  if(!check_resources(s, pr, report)) {
	
	    switch(pr->action) {
	    case ACTION_ALERT:
	      if(s->def_timeout) s->nstart++;
	      alert_resource(s, "Reason: %s\n", report);
	      /* We are also interested in other alerts/stops/restarts! */
	      pr->cycle=0;
	      break;

	    case ACTION_STOP:
	      do_stop(s, "Reason: %s\n", report);
	      reset_resource_counter(s);
	      goto reinstall;

	    case ACTION_RESTART:
	      do_restart(s, "Reason: %s\n", report);
	      reset_resource_counter(s);
	      goto reinstall;

	    default:
	      log("'%s' Unknow resource failure action.\n", s->name);
	      break;
	    }

	  }
      
        }
	
      } else {

        log("'%s' failed to get service data\n", s->name);
      
      }

    }

    /* Test each host:port and protocol in the service's portlist */
    for(pp= s->portlist; pp; pp= pp->next) {
    
      if(!check_connection(s, pp, report)) {
      
        do_restart(s, "Reason: %s\n", report);
        goto reinstall;
      
      }
    
    }
    break;

  case TASK_DEVICE:

    /* Test devices */
    if(s->devicelist) {

      if(!get_fsusage(s->path, s->devinfo)) {
        do_restart(s, "Reason: unable to read device state\n");
        goto reinstall;
      }

      for(td= s->devicelist; td; td= td->next) {

        if(!check_device(s, td, report)) {

          switch(td->action) {
          case ACTION_ALERT:
            alert_resource(s, "Reason: %s\n", report);
            break;

          case ACTION_STOP:
            do_stop(s, "Reason: %s\n", report);
            goto reinstall;

	  case ACTION_RESTART:
	    do_restart(s, "Reason: %s\n", report);
	    goto reinstall;

          default:
            log("'%s' Unknow device failure action.\n", s->name);
            break;
          }

        }

      }

    }
    break;

  case TASK_FILE:

    if(stat(s->path, &stat_buf) != 0) {
      log("'%s' doesn't exist\n", s->name);
      do_restart(s, "Reason: file doesn't exist\n");
      goto reinstall;
    }

    if(!S_ISREG(stat_buf.st_mode)) {
      log("'%s' is not regular file\n", s->name);
      do_restart(s, "Reason: object is not file\n");
      goto reinstall;
    }

    if( s->perm != -1 && (stat_buf.st_mode & 07777) != s->perm ) {
      log("'%s' permission test failed -- current permission: %o\n",
        s->name, stat_buf.st_mode&07777);
      do_restart(s, "Reason: permission test failed: %o\n",
		 stat_buf.st_mode&07777);
      goto reinstall;
    }

    DEBUG("'%s' file permission check passed [current permission=%o]\n",
      s->name, stat_buf.st_mode&07777);

    break;

  case TASK_DIRECTORY:

    if(stat(s->path, &stat_buf) != 0) {
      log("'%s' doesn't exist\n", s->name);
      do_restart(s, "Reason: directory doesn't exist\n");
      goto reinstall;
    }

    if(!S_ISDIR(stat_buf.st_mode)) {
      log("'%s' is not directory\n", s->name);
      do_restart(s, "Reason: object is not directory\n");
      goto reinstall;
    }

    if( s->perm != -1 && (stat_buf.st_mode & 07777) != s->perm ) {
      log("'%s' permission test failed -- current permission: %o\n",
        s->name, stat_buf.st_mode&07777);
      do_restart(s, "Reason: permission test failed: %o\n",
		 stat_buf.st_mode&07777);
      goto reinstall;
    }

    DEBUG("'%s' directory permission check passed [current permission=%o]\n",
      s->name, stat_buf.st_mode&07777);

    break;

  default:
    break;

  }


  /** Common task actions */
  for(tl= s->timestamplist; tl; tl= tl->next) {
      
    if(!check_timestamp(s, tl, report)) {
	
      switch(tl->action) {
      case ACTION_ALERT:
          if(s->def_timeout) s->nstart++;
	  alert_timestamp(s, "Reason: %s\n", report);
	  break; /* continue */
	  
      case ACTION_STOP:
	  do_stop(s, "Reason: %s\n", report);
	  goto reinstall;

      case ACTION_RESTART:
	  do_restart(s, "Reason: %s\n", report);
	  goto reinstall;

      default:
	  log("'%s' Unknow timestamp failure action.\n", s->name);
	  break;

      }

    }

  } 	
      
  reinstall:
  
  /* Remove the SIGTERM block */
  pthread_sigmask(SIG_SETMASK, &os, NULL);

  /* Release service action mutex */
  END_LOCK;
 
}


/**
 * Restart the service if the start and stop
 * program is defined, and validate is TRUE.
 */
static void do_restart(Service_T s, char *m, ...) {

  va_list ap;
  char *tmp = NULL;

  ASSERT(s);
  
  if(!s->do_validate)
    return;

  if(s->def_timeout)
    s->nstart++;

  va_start(ap, m);
  if(m)
    tmp=format(m, ap);
  va_end(ap);

  if(s->mode!=MODE_PASSIVE && s->start && s->stop) {

    log("Trying to restart '%s'\n", s->name);
    check_service(s->name, "restart");
    alert_restart(s, "%s", tmp);

  } else {

    alert_failed(s, "%s", tmp);

  }

  free(tmp);

}


/**
 * Start the service if the start program is defined,
 * and validate is TRUE.
 */
static void do_start(Service_T s, char *m, ...) {

  va_list ap;
  char *tmp = NULL;

  ASSERT(s);

  if(!s->do_validate)
    return;

  if(s->def_timeout)
    s->nstart++;

  va_start(ap, m);
  if(m)
    tmp=format(m, ap);
  va_end(ap);

  if(s->mode!= MODE_PASSIVE && s->start) {

    check_service(s->name, "start");
    alert_restart(s, "%s", tmp);

  } else {

    alert_failed(s, "%s", tmp);

  }

  free(tmp);
  
}


/**
 * Stop the service if the stop program is defined,
 * and validate is TRUE.
 */
static void do_stop(Service_T s, char *m, ...) {

  va_list ap;
  char *tmp = NULL;

  ASSERT(s);

  if(!s->do_validate)
    return;

  va_start(ap, m);
  if(m)
    tmp=format(m, ap);
  va_end(ap);

  LOCK(Run.mutex)
    s->do_validate= FALSE;
  END_LOCK;

  if(s->mode!= MODE_PASSIVE && s->stop) {

    alert_stop(s, "%s", tmp);
    check_service(s->name, "stop");

  } else {

    if(s->def_timeout)
      s->nstart++;

    alert_failed(s, "%s", tmp);

  }

  free(tmp);

}


/**
 * Returns the logical OR of pre-condition tests.
 * A variant of the template-method pattern
 */
static int do_not_validate(Service_T s) {

  ASSERT(s);
  
  return(!s->do_validate  ||
	 check_skip(s)    ||
	 check_timeout(s) ||
	 check_checksum(s));

}


/**
 * Returns TRUE if the connection and protocol test succeeded
 * otherwise FALSE.
 */
static int check_connection(Service_T s, Port_T pp, char *report) {

  volatile int rv= TRUE;

  ASSERT(s && pp);
  
  /* Control comes here if a timeout occures */
  if(sigsetjmp(timeout, TRUE)) {
    
    snprintf(report, STRLEN, 
	     "timed out when testing %s [%s]",
	     pp->address, pp->protocol->name);
    
    log("'%s' %s\n", s->name, report); 

    rv= FALSE;
    goto error;
    
  }

  /* Set a timeout handler and activate the timer */
  set_alarm_handler(connection_timeout);
  alarm(CHECK_TIMEOUT);

  /* Open a socket to the destination INET[hostname:port] or UNIX[pathname] */
  if((pp->socket= create_generic_socket(pp)) < 0) {
      
    snprintf(report, STRLEN, 
	     "does not accept connection at %s.",
	     pp->address);

    log("'%s' %s\n", s->name, report); 

    rv= FALSE;
    goto error;
    
  } else {
    
    DEBUG("'%s' succeeded connecting to %s\n", s->name, pp->address);
    
  }

  /* Verify that the socket is ready for i|o */
  if(!check_connection_io(pp)) {
    
    snprintf(report, STRLEN, 
	     "socket at %s is not ready for i|o -- %s",
	     pp->address, STRERROR);
    
    log("'%s' %s\n", s->name, report); 

    rv= FALSE;
    goto error;

  }

  if(pp->ssl != NULL) {

    if(!embed_ssl_socket(pp->ssl, pp->socket)) {

      snprintf(report, STRLEN, 
	       "failed establish SSL communication on socket at %s",
	       pp->address);
      
      log("'%s' %s\n", s->name, report); 
      rv= FALSE;
      goto error;
      
    }

    if(pp->certmd5 != NULL) {

      if(! check_ssl_md5sum(pp->ssl, pp->certmd5)) {

	snprintf(report, STRLEN, 
		 "md5sums of SSL certificates do not match at %s",
		 pp->address);
	
	log("'%s' %s\n", s->name, report); 
	rv= FALSE;
	goto error;

      }

    }

  }

  /* Run the protocol verification routine through the socket */
  if(! pp->protocol->check(pp)) {
    
    snprintf(report, STRLEN, 
	     "test with protocol [%s] failed at %s.",
	     pp->protocol->name, pp->address);

    log("'%s' %s\n", s->name, report); 
    
    rv= FALSE;
    goto error;
    
  } else {
    
    DEBUG("'%s' succeeded testing protocol [%s] at %s\n",
      s->name, pp->protocol->name, pp->address);
  }

  error:

  alarm(0);

  if(pp->ssl != NULL) {

    cleanup_ssl_socket(pp->ssl);
    
  }

  close_socket(pp->socket);
  
  return rv;
      
}


/**
 * Returns TRUE if the process state (e.g. Zombie) test succeeded
 * otherwise FALSE.
 */
static int check_process_state(Service_T s, char *report) {

  ProcInfo_T pi;

  ASSERT(s);
  
  pi= s->procinfo;
  
  if(pi->status_flag & PROCESS_ZOMBIE) {

    snprintf(report, STRLEN, 
	     "process with pid %d is a zombie\n", pi->pid);

    log("'%s' %s\n", s->name, report); 

    /* We do not check the process anymore if it's a zombie
     since such a process is (usually) unmanageable */
    LOCK(Run.mutex)
	s->do_validate= FALSE;
    END_LOCK;
    
    return FALSE;

  } else {

    DEBUG("'%s' zombie check passed [status_flag=%04x]\n",
      s->name,  pi->status_flag);

  }

  return TRUE;

}


/**
 * Resets the resource counter
 */
static void reset_resource_counter(Service_T s) {
  
  Resource_T pr;

  ASSERT(s);
  
  for(pr= s->resourcelist; pr; pr= pr->next) {
    
      pr->cycle=0;

  }
  
}


/**
 * Returns TRUE if the proc test succeeded
 * otherwise FALSE.
 */
static int check_resources(Service_T s, Resource_T pr, char *report) {

  ProcInfo_T pi;
  int okay= TRUE;

  ASSERT(s);
  ASSERT(pr);

  pi= s->procinfo;
  
  switch(pr->resource_id) {
  case RESOURCE_ID_CPU_PERCENT: 
    if(compare_value(pr->operator, pi->cpu_percent, pr->limit)) {

      vlog(report, STRLEN, s, 
	   "cpu usage of %.1f%% matches resource limit [cpu usage%s%.1f%%]",
	   pi->cpu_percent/10.0, operatorshortnames[pr->operator],
	   pr->limit/10.0);

      okay= FALSE;

    } else {

      DEBUG("'%s' cpu usage check passed [current cpu usage=%.1f%%]\n", 
	s->name, pi->cpu_percent/10.0);

    }
    break;

  case RESOURCE_ID_MEM_PERCENT: 
    if(compare_value(pr->operator, pi->mem_percent, pr->limit)) {

      vlog(report, STRLEN, s, 
	   "mem usage of %.1f%% matches resource limit [mem usage%s%.1f%%]",
	   pi->mem_percent/10.0, operatorshortnames[pr->operator],
	   pr->limit/10.0);      

      okay= FALSE;

    } else {
      
	
      DEBUG("'%s' mem usage check passed [current mem usage=%.1f%%]\n", 
        s->name, pi->mem_percent/10.0);
	
    }
    break;

  case RESOURCE_ID_MEM_KBYTE: 

    if(compare_value(pr->operator, pi->mem_kbyte, pr->limit)) {

      vlog(report, STRLEN, s, 
	   "mem amount of %ldkB matches resource limit [mem amount%s%ldkB]", 
	   pi->mem_kbyte, operatorshortnames[pr->operator],
	   pr->limit);      
      
      okay= FALSE;
    
    } else {

	
      DEBUG("'%s' mem amount check passed [current mem amount=%ldkB]\n", 
	s->name, pi->mem_kbyte);
      
    }
    break;

  case RESOURCE_ID_LOAD1: 

    if(compare_value(pr->operator, (int)(Run.loadavg[0]*10.0), pr->limit)) {

      vlog(report, STRLEN, s, 
	   "loadavg(1min) of %.1f matches resource limit "
	   "[loadavg(1min)%s%.1f]", 
	   Run.loadavg[0], operatorshortnames[pr->operator],
	   pr->limit/10.0);      
      
      okay= FALSE;
    
    } else {

	
      DEBUG("'%s' loadavg(1min) check passed [current loadavg(1min)=%.1f]\n", 
	s->name, Run.loadavg[0]);
      
    }
    break;

  case RESOURCE_ID_LOAD5: 

    if(compare_value(pr->operator, (int)(Run.loadavg[1]*10.0), pr->limit)) {

      vlog(report, STRLEN, s, 
	   "loadavg(5min) of %.1f matches resource limit "
	   "[loadavg(5min)%s%.1f]", 
	   Run.loadavg[1], operatorshortnames[pr->operator],
	   pr->limit/10.0);      
      
      okay= FALSE;
    
    } else {

	
      DEBUG("'%s' loadavg(5min) check passed [current loadavg(5min)=%.1f]\n", 
	s->name, Run.loadavg[1]);
      
    }
    break;

  case RESOURCE_ID_LOAD15: 

    if(compare_value(pr->operator, (int)(Run.loadavg[2]*10.0), pr->limit)) {

      vlog(report, STRLEN, s, 
	   "loadavg(15min) of %.1f matches resource limit "
	   "[loadavg(15min)%s%.1f]", 
	   Run.loadavg[2], operatorshortnames[pr->operator],
	   pr->limit/10.0);      
      
      okay= FALSE;
    
    } else {

      DEBUG("'%s' loadavg(15min) check passed "
	    "[current loadavg(15min)=%.1f]\n", 
	    s->name, Run.loadavg[2]);
      
    }
    break;

  case RESOURCE_ID_CHILDREN:

    if(compare_value(pr->operator, pi->children, pr->limit)) {

      vlog(report, STRLEN, s,
        "children of %i matches resource limit "
        "[children%s%i]",
        pi->children, operatorshortnames[pr->operator],
        pr->limit);

      okay= FALSE;

    } else {

      DEBUG("'%s' children check passed "
        "[current children=%i]\n",
        s->name, pi->children);

    }
    break;

  case RESOURCE_ID_TOTAL_MEM_KBYTE:

    if(compare_value(pr->operator, pi->total_mem_kbyte, pr->limit)) {

      vlog(report, STRLEN, s,
          "total mem amount of %ldkB matches resource limit"
          " [total mem amount%s%ldkB]",
          pi->total_mem_kbyte, operatorshortnames[pr->operator],
          pr->limit);

      okay= FALSE;

    } else {

      DEBUG("'%s' total mem amount check passed"
        " [current total mem amount=%ldkB]\n",
        s->name, pi->total_mem_kbyte);

    }
    break;

  default:

      log("'%s' error: unknow resource ID: [%d]\n", s->name, pr->resource_id);

  }

  if(okay && pr->cycle > 0) {

    pr->cycle--;

  } else if(! okay) {

    pr->cycle++;

  }

  if(pr->cycle >= pr->max_cycle) {

    return FALSE;

  }
    
  return TRUE;
  
}


/**
 * Returns TRUE if the service timed out, otherwise FALSE. 
 */
static int check_timeout(Service_T s) {

  ASSERT(s);
  
  if(!s->def_timeout) {

    return FALSE;

  }
  
  /*
   * Start counting cycles
   */
  if(s->nstart > 0) {
    
    s->ncycle++;
    
  }
  
  /*
   * Check timeout
   */
  if(s->nstart >= s->to_start && s->ncycle <= s->to_cycle) {
    
    /*
     * Got timeout - Set do not validate flag. This statement is
     * synchronized since the same data can be set from the http
     * thread
     */
    LOCK(Run.mutex)
	s->do_validate= FALSE;
    END_LOCK;
    
    /*
     * Log and notify the user that a timeout occured
     */
    log("**Alert** service '%s' timed out and will not be checked anymore.\n",
	s->name);
      
    alert_timeout(s, NULL);
    
    return TRUE;
    
  }
  
  /*
   * Stop counting and reset if the
   * cycle interval is passed
   */
  if(s->ncycle > s->to_cycle) {
    
    s->ncycle= 0;
    s->nstart= 0;
    
  }

  return FALSE;
  
}


/**
 * Returns TRUE if validation should be skiped for
 * this service in this cycle, otherwise FALSE
 */
static int check_skip(Service_T s) {

  ASSERT(s);
  
  if(!s->def_every) {

    return FALSE;

  }
  
  if(++s->nevery < s->every) {
    
      return TRUE;

  }
  
  s->nevery= 0;

  return FALSE;

}


/**
 * Returns TRUE if the checksum was changed for any associated
 * programs and send an alert warning. Also make sure that the program
 * will not be validated, started or restarted again in this
 * session. Returns FALSE if the checksum is the same or not defined
 * for this service.
 */
static int check_checksum(Service_T s) {

  Checksum_T c;

  ASSERT(s);

  if(!s->def_checksum) {

    return FALSE;

  }

  for(c= s->checksumlist; c; c= c->next) {
    
    if(c->file && c->md5 && !check_md5(c->file, c->md5)) {

      log("'%s' CHECKSUM ERROR for %s\n", s->name, c->file);
      alert_checksum(s, NULL);
      
      LOCK(Run.mutex)
	  s->do_validate= FALSE;
          s->has_checksum_error= TRUE;
      END_LOCK;
      
      return TRUE;

    }
    
  }
  
  DEBUG("'%s' have valid checksums\n", s->name);
  
  return FALSE;

}


/**
 * Returns TRUE if the timestamp test succeded, otherwise FALSE
 */
static int check_timestamp(Service_T s, Timestamp_T t, char *report) {

  time_t now;
  time_t timestamp;

  ASSERT(s);
  ASSERT(t);

  if((int)time(&now) == -1) {
    vlog(report, STRLEN, s, "can't get actual time");      
    return FALSE;
  }

  if(!(timestamp= get_timestamp(t->pathname, S_IFDIR|S_IFREG))) {
    vlog(report, STRLEN, s, "can't get timestamp for %s", t->pathname);      
    return FALSE;
  }

  if(compare_value(t->operator, (int)(now - timestamp), t->time)) {
    vlog(report, STRLEN, s, "timestamp test failed for %s", t->pathname);      
    return FALSE;
  }

  DEBUG("'%s' timestamp test passed for %s\n", s->name, t->pathname); 
      
  return TRUE;

}


/**
 * Returns TRUE if the device test succeeded
 * otherwise FALSE.
 */
static int check_device(Service_T t, Device_T td, char *report) {

  ASSERT(t);
  ASSERT(td);

  if( (td->limit_percent < 0) && (td->limit_absolute < 0) ) {

    log("'%s' error: device limit not set\n", t->name);
    return FALSE;

  }

  switch(td->resource) {

  case RESOURCE_ID_INODE:

    if(t->devinfo->f_files <= 0) {

      DEBUG("'%s' filesystem doesn't support inodes\n", t->name);
      return TRUE;

    }

    if(td->limit_percent >= 0) {

      if(compare_value(
           td->operator,
           100 * (t->devinfo->f_files - t->devinfo->f_filesfree) /
	   t->devinfo->f_files, td->limit_percent)) {

        vlog(report, STRLEN, t,
          "inode usage %d%% matches resource limit [inode usage%s%d%%]",
          100 * (t->devinfo->f_files - t->devinfo->f_filesfree) /
	     t->devinfo->f_files,
          operatorshortnames[td->operator],
          td->limit_percent);

        return FALSE;

      }

    } else {

      if(compare_value(td->operator, t->devinfo->f_files -
		       t->devinfo->f_filesfree, td->limit_absolute)) {

        vlog(report, STRLEN, t,
          "inode usage %d matches resource limit [inode usage%s%d]",
          t->devinfo->f_files - t->devinfo->f_filesfree,
          operatorshortnames[td->operator],
          td->limit_absolute);

        return FALSE;

      }

    }

    DEBUG("'%s' inode usage check passed [current inode usage=%.1f%%]\n",
      t->name,
      (float) 100 * (t->devinfo->f_files - t->devinfo->f_filesfree) /
	  t->devinfo->f_files);

    return TRUE;

  case RESOURCE_ID_SPACE:

      if(td->limit_percent >= 0) {

        if(compare_value(
             td->operator,
             100 * (t->devinfo->f_blocks - t->devinfo->f_blocksfree) /
	     t->devinfo->f_blocks, td->limit_percent)) {

          vlog(report, STRLEN, t,
	       "space usage %d%% matches resource limit [space usage%s%d%%]",
            100 * (t->devinfo->f_blocks - t->devinfo->f_blocksfree) /
	       t->devinfo->f_blocks, operatorshortnames[td->operator],
	       td->limit_percent);

          return FALSE;

        }

      } else {

        if(compare_value(td->operator, t->devinfo->f_blocks -
			 t->devinfo->f_blocksfree, td->limit_absolute)) {

          vlog(report, STRLEN, t,
       "space usage %d blocks matches resource limit [space usage%s%d blocks]",
            t->devinfo->f_blocks - t->devinfo->f_blocksfree,
            operatorshortnames[td->operator],
            td->limit_absolute);

          return FALSE;

        }

      }

    DEBUG("'%s' space usage check passed [current space usage=%.1f%%]\n",
      t->name,
      (float) 100 * (t->devinfo->f_blocks - t->devinfo->f_blocksfree) /
	  t->devinfo->f_blocks);

    return TRUE;

  default:

    log("'%s' error: unknow resource type: [%d]\n", t->name, td->resource);
    return FALSE;

  }

}


/**
 * Signal handler for connection timeout
 */
static void connection_timeout(int sig) {
  
  siglongjmp(timeout, TRUE);
  
}


/**
 * Logging with report generation
 */
static void vlog(char * report, int n, Service_T s, char *m,...) {

  va_list ap;
  char *tmp = NULL;

  va_start(ap, m);

  if(m) { 

    tmp=format(m,ap);

  }

  va_end(ap);

  strncpy(report, tmp, n);
  log("'%s' %s\n", s->name, report); 

  free(tmp);
}


/**
 * Comparison of values. Returns TRUE if comparison matches, otherwise
 * FALSE.
 */
static int compare_value(int operator, int left, int right) {

    switch(operator) {

    case OPERATOR_GREATER:

      if(left > right)
	return TRUE;
      break;

    case OPERATOR_LESS:

      if(left < right)
	return TRUE;
      break;

    case OPERATOR_EQUAL:

      if(left == right)
	return TRUE;
      break;

    case OPERATOR_NOTEQUAL:

      if(left != right)
	return TRUE;
      break;

    default:
      log("Unknow comparison operator\n");
      return FALSE;

    }

    return FALSE;

}

